# -*- coding: utf-8 -*-
import json
import tempfile
import unittest
import uuid
from collections import Counter
from io import StringIO
from pathlib import Path

import numpy as np
from hocr_spec import HocrValidator
from lxml import etree

from kraken import containers, serialization
from kraken.lib import xml

thisfile = Path(__file__).resolve().parent
resources = thisfile / 'resources'

def roundtrip(self, records, fp):
    """
    Checks that the order of lines after serialization and deserialization is
    equal to the records.
    """
    with tempfile.NamedTemporaryFile() as out:
        fp.seek(0)
        out.write(fp.getvalue().encode('utf-8'))
        doc = xml.XMLPage(out.name).to_container().lines
        for orig_line, parsed_line in zip(records, doc):
            self.assertSequenceEqual(np.array(orig_line.baseline).tolist(),
                                     np.array(parsed_line.baseline).tolist(),
                                     msg='Baselines differ after serialization.')

def validate_hocr(self, fp):
    fp.seek(0)

    validator = HocrValidator('standard')
    report = validator.validate(fp, parse_strict=True)
    self.assertTrue(report.is_valid())

    doc = etree.fromstring(fp.getvalue().encode('utf-8'))

    ids = [x.get('id') for x in doc.findall('.//*[@id]')]
    counts = Counter(ids)
    self.assertEqual(counts.most_common(1)[0][1], 1, msg='Duplicate IDs in hOCR output')

def validate_page(self, fp):
    doc = etree.fromstring(fp.getvalue().encode('utf-8'))

    ids = [x.get('id') for x in doc.findall('.//*[@id]')]
    counts = Counter(ids)
    if len(counts):
        self.assertEqual(counts.most_common(1)[0][1], 1, msg='Duplicate IDs in PageXML output')

    with open(resources / 'pagecontent.xsd') as schema_fp:
        page_schema = etree.XMLSchema(etree.parse(schema_fp))
        page_schema.assertValid(doc)

def validate_alto(self, fp):
    doc = etree.fromstring(fp.getvalue().encode('utf-8'))

    ids = [x.get('ID') for x in doc.findall('.//*[@ID]')]
    counts = Counter(ids)
    self.assertEqual(counts.most_common(1)[0][1], 1, msg='Duplicate IDs in ALTO output')

    with open(resources / 'alto-4-3.xsd') as schema_fp:
        alto_schema = etree.XMLSchema(etree.parse(schema_fp))
        alto_schema.assertValid(doc)


class TestSerializations(unittest.TestCase):
    """
    Tests for output serialization
    """
    def setUp(self):
        with open(resources /'records.json', 'r') as fp:
            self.box_records = [containers.BBoxOCRRecord(**x) for x in json.load(fp)]

        with open(resources / 'bl_records.json', 'r') as fp:
            recs = json.load(fp)
            self.bl_records = [containers.BaselineOCRRecord(**bl) for bl in recs['lines']]
            self.bl_regions = recs['regions']

        self.box_segmentation = containers.Segmentation(type='bbox',
                                                        imagename='foo.png',
                                                        text_direction='horizontal-lr',
                                                        lines=self.box_records,
                                                        script_detection=True,
                                                        regions={})

        self.bl_segmentation = containers.Segmentation(type='baselines',
                                                       imagename='foo.png',
                                                       text_direction='horizontal-lr',
                                                       lines=self.bl_records,
                                                       script_detection=True,
                                                       regions={})

        self.bl_segmentation_regs = containers.Segmentation(type='baselines',
                                                            imagename='foo.png',
                                                            text_direction='horizontal-lr',
                                                            lines=self.bl_records,
                                                            script_detection=True,
                                                            regions=self.bl_regions)

        self.bl_seg_nolines_regs = containers.Segmentation(type='baselines',
                                                           imagename='foo.png',
                                                           text_direction='horizontal-lr',
                                                           script_detection=False,
                                                           lines=[],
                                                           regions=self.bl_regions)

        self.metadata_steps = [containers.ProcessingStep(id=str(uuid.uuid4()),
                                                         category='preprocessing',
                                                         description='PDF image extraction',
                                                         settings={}),
                               containers.ProcessingStep(id=str(uuid.uuid4()),
                                                         category='processing',
                                                         description='Baseline and region segmentation',
                                                         settings={'model': 'foo.mlmodel', 'text_direction': 'horizontal-lr'}),
                               containers.ProcessingStep(id=str(uuid.uuid4()),
                                                         category='processing',
                                                         description='Text line recognition',
                                                         settings={'text_direction': 'horizontal-lr',
                                                                   'models': 'bar.mlmodel',
                                                                   'pad': 16,
                                                                   'bidi_reordering': True})]


    def test_box_vertical_hocr_serialization(self):
        """
        Test vertical line hOCR serialization
        """
        fp = StringIO()

        fp.write(serialization.serialize(self.box_segmentation, writing_mode='vertical-lr', template='hocr'))
        validate_hocr(self, fp)

    def test_box_hocr_serialization(self):
        """
        Test hOCR serialization
        """
        fp = StringIO()

        fp.write(serialization.serialize(self.box_segmentation, template='hocr'))
        validate_hocr(self, fp)

    def test_box_alto_serialization_validation(self):
        """
        Validates output against ALTO schema
        """
        fp = StringIO()

        fp.write(serialization.serialize(self.box_segmentation, template='alto'))
        validate_alto(self, fp)

    def test_box_abbyyxml_serialization_validation(self):
        """
        Validates output against abbyyXML schema
        """
        fp = StringIO()

        fp.write(serialization.serialize(self.box_segmentation, template='abbyyxml'))
        doc = etree.fromstring(fp.getvalue().encode('utf-8'))
        with open(resources / 'FineReader10-schema-v1.xml') as schema_fp:
            abbyy_schema = etree.XMLSchema(etree.parse(schema_fp))
            abbyy_schema.assertValid(doc)

    def test_box_pagexml_serialization_validation(self):
        """
        Validates output against abbyyXML schema
        """
        fp = StringIO()

        fp.write(serialization.serialize(self.box_segmentation, template='pagexml'))
        validate_page(self, fp)

    def test_bl_alto_serialization_validation(self):
        """
        Validates output against ALTO schema
        """
        fp = StringIO()

        fp.write(serialization.serialize(self.bl_segmentation, template='alto'))
        validate_alto(self, fp)
        roundtrip(self, self.bl_records, fp)

    def test_bl_abbyyxml_serialization_validation(self):
        """
        Validates output against abbyyXML schema
        """
        fp = StringIO()

        fp.write(serialization.serialize(self.bl_segmentation, template='abbyyxml'))
        doc = etree.fromstring(fp.getvalue().encode('utf-8'))
        with open(resources / 'FineReader10-schema-v1.xml') as schema_fp:
            abbyy_schema = etree.XMLSchema(etree.parse(schema_fp))
            abbyy_schema.assertValid(doc)

    def test_bl_pagexml_serialization_validation(self):
        """
        Validates output against PageXML schema
        """
        fp = StringIO()

        fp.write(serialization.serialize(self.bl_segmentation, template='pagexml'))
        validate_page(self, fp)
        roundtrip(self, self.bl_records, fp)

    def test_bl_region_alto_serialization_validation(self):
        """
        Validates output against ALTO schema
        """
        fp = StringIO()

        fp.write(serialization.serialize(self.bl_segmentation_regs, template='alto'))
        validate_alto(self, fp)
        roundtrip(self, self.bl_records, fp)

    def test_bl_region_abbyyxml_serialization_validation(self):
        """
        Validates output against abbyyXML schema
        """
        fp = StringIO()

        fp.write(serialization.serialize(self.bl_segmentation_regs, template='abbyyxml'))
        doc = etree.fromstring(fp.getvalue().encode('utf-8'))
        with open(resources / 'FineReader10-schema-v1.xml') as schema_fp:
            abbyy_schema = etree.XMLSchema(etree.parse(schema_fp))
            abbyy_schema.assertValid(doc)

    def test_bl_region_pagexml_serialization_validation(self):
        """
        Validates output against PageXML schema
        """
        fp = StringIO()

        fp.write(serialization.serialize(self.bl_segmentation_regs, template='pagexml'))
        validate_page(self, fp)
        roundtrip(self, self.bl_records, fp)

    def test_region_only_alto_serialization_validation(self):
        """
        Validates output without baselines (but regions) against ALTO schema
        """
        fp = StringIO()

        fp.write(serialization.serialize(self.bl_seg_nolines_regs, template='alto'))
        validate_alto(self, fp)

    def test_region_only_abbyyxml_serialization_validation(self):
        """
        Validates output without baselines (but regions) against abbyyXML schema
        """
        fp = StringIO()

        fp.write(serialization.serialize(self.bl_seg_nolines_regs, template='abbyyxml'))
        doc = etree.fromstring(fp.getvalue().encode('utf-8'))
        with open(resources / 'FineReader10-schema-v1.xml') as schema_fp:
            abbyy_schema = etree.XMLSchema(etree.parse(schema_fp))
            abbyy_schema.assertValid(doc)

    def test_region_only_pagexml_serialization_validation(self):
        """
        Validates output without baselines (but regions) against PageXML schema
        """
        fp = StringIO()

        fp.write(serialization.serialize(self.bl_seg_nolines_regs, template='pagexml'))
        validate_page(self, fp)

    def test_bl_region_alto_serialization_validation_steps(self):
        """
        Validates output with processing steps against ALTO schema
        """
        fp = StringIO()

        fp.write(serialization.serialize(self.bl_segmentation, template='alto', processing_steps=self.metadata_steps))

        validate_alto(self, fp)
        roundtrip(self, self.bl_records, fp)

    def test_bl_region_abbyyxml_serialization_validation_steps(self):
        """
        Validates output with processing steps against abbyyXML schema
        """
        fp = StringIO()

        fp.write(serialization.serialize(self.bl_segmentation, template='abbyyxml', processing_steps=self.metadata_steps))

        doc = etree.fromstring(fp.getvalue().encode('utf-8'))
        with open(resources / 'FineReader10-schema-v1.xml') as schema_fp:
            abbyy_schema = etree.XMLSchema(etree.parse(schema_fp))
            abbyy_schema.assertValid(doc)

    def test_bl_region_pagexml_serialization_validation_steps(self):
        """
        Validates output with processing steps against PageXML schema
        """
        fp = StringIO()
        fp.write(serialization.serialize(self.bl_segmentation, template='pagexml', processing_steps=self.metadata_steps))
        validate_page(self, fp)
        roundtrip(self, self.bl_records, fp)




"""
Layers for VGSL models
"""
from typing import TYPE_CHECKING, Tuple

import torch
from torch import nn

if TYPE_CHECKING:
    from kraken.lib.vgsl import VGSLBlock

# all tensors are ordered NCHW, the "feature" dimension is C, so the output of
# an LSTM will be put into C same as the filters of a CNN.

__all__ = ['MLP']


class MLP(nn.Module):
    """
    A simple 2 layer MLP for reading order determination.
    """
    def __init__(self, feature_size: int, hidden_size: int):
        super(MLP, self).__init__()
        self.fc1 = nn.Linear(feature_size, hidden_size)
        self.relu = nn.ReLU()
        self.fc2 = nn.Linear(hidden_size, 1)
        self.feature_size = feature_size
        self.hidden_size = hidden_size
        self.class_mapping = None

    def forward(self, x):
        x = self.fc1(x)
        x = self.relu(x)
        return self.fc2(x)

    def get_shape(self, input: Tuple[int, int, int, int]) -> Tuple[int, int, int, int]:
        """
        Calculates the output shape from input 4D tuple NCHW.
        """
        return input

    def get_spec(self, name) -> 'VGSLBlock':
        """
        Generates a VGSL spec block from the layer instance.
        """
        return f'[1,0,0,1 RO{{{name}}}{self.feature_size},{self.hidden_size}]'

    def deserialize(self, name: str, spec) -> None:
        """
        Sets the weights of an initialized module from a CoreML protobuf spec.
        """
        # extract 1st linear projection parameters
        lin = [x for x in spec.neuralNetwork.layers if x.name == '{}_mlp_lin_0'.format(name)][0].innerProduct
        weights = torch.Tensor(lin.weights.floatValue).resize_as_(self.fc1.weight.data)
        bias = torch.Tensor(lin.bias.floatValue)
        self.fc1.weight = torch.nn.Parameter(weights)
        self.fc1.bias = torch.nn.Parameter(bias)
        # extract 2nd linear projection parameters
        lin = [x for x in spec.neuralNetwork.layers if x.name == '{}_mlp_lin_1'.format(name)][0].innerProduct
        weights = torch.Tensor(lin.weights.floatValue).resize_as_(self.fc2.weight.data)
        bias = torch.Tensor(lin.bias.floatValue)
        self.fc2.weight = torch.nn.Parameter(weights)
        self.fc2.bias = torch.nn.Parameter(bias)

    def serialize(self, name: str, input: str, builder):
        """
        Serializes the module using a NeuralNetworkBuilder.
        """
        builder.add_inner_product(f'{name}_mlp_lin_0', self.fc1.weight.data.numpy(),
                                  self.fc1.bias.data.numpy(),
                                  self.feature_size, self.hidden_size,
                                  has_bias=True, input_name=input, output_name=f'{name}_mlp_lin_0')
        builder.add_activation(f'{name}_mlp_lin_0_relu', 'RELU', f'{name}_mlp_lin_0', f'{name}_mlp_lin_0_relu')
        builder.add_inner_product(f'{name}_mlp_lin_1', self.fc2.weight.data.numpy(),
                                  self.fc2.bias.data.numpy(),
                                  self.hidden_size, 1,
                                  has_bias=True, input_name=f'{name}_mlp_lin_0_relu', output_name=f'{name}_mlp_lin_1')
        return name



# -*- coding: utf-8 -*-

import json
import unittest
from pathlib import Path

from pytest import raises

import kraken
from kraken.lib import xml
from kraken.lib.exceptions import KrakenInputException
from kraken.lib.train import KrakenTrainer, RecognitionModel, SegmentationModel

thisfile = Path(__file__).resolve().parent
resources = thisfile / 'resources'

class TestKrakenTrainer(unittest.TestCase):
    """
    Tests for KrakenTrainer class
    """
    def setUp(self):
        self.xml = resources / '170025120000003,0074.xml'
        self.bls = xml.XMLPage(self.xml)
        self.box_lines = [resources / '000236.png']
        self.model = resources / 'model_small.mlmodel'

    def test_krakentrainer_rec_box_load_fail(self):
        training_data = self.box_lines
        evaluation_data = self.box_lines
        module = RecognitionModel(format_type='path',
                                  model=self.model,
                                  training_data=training_data,
                                  evaluation_data=evaluation_data,
                                  resize='fail')
        with raises(KrakenInputException):
            module.setup()

    def test_krakentrainer_rec_bl_load_fail(self):
        """
        Tests that the proper exception is raised when loading model not fitting the dataset.
        """
        training_data = [self.xml]
        evaluation_data = [self.xml]
        module = RecognitionModel(format_type='xml',
                                  model=self.model,
                                  training_data=training_data,
                                  evaluation_data=evaluation_data,
                                  resize='fail')
        with raises(KrakenInputException):
            module.setup()

    def test_krakentrainer_rec_box_load_union(self):
        """
        Tests that adaptation works in `union` mode.

        The dataset brings 15 new characters. Input model had 3. There is one spec. char to account for.
        """
        training_data = self.box_lines
        evaluation_data = self.box_lines
        module = RecognitionModel(format_type='path',
                                  model=self.model,
                                  training_data=training_data,
                                  evaluation_data=evaluation_data,
                                  resize='union')
        module.setup("fit")
        self.assertEqual(module.nn.seg_type, 'bbox')
        self.assertIsInstance(module.train_set.dataset, kraken.lib.dataset.GroundTruthDataset)
        trainer = KrakenTrainer(max_steps=1)
        self.assertEqual(module.nn.named_spec[-1].split("c")[-1], '19')

    def test_krakentrainer_rec_box_load_new(self):
        """
        Tests that adaptation works in `new` mode.
        """
        training_data = self.box_lines
        evaluation_data = self.box_lines
        module = RecognitionModel(format_type='path',
                                  model=self.model,
                                  training_data=training_data,
                                  evaluation_data=evaluation_data,
                                  resize='new')
        module.setup("fit")
        self.assertEqual(module.nn.seg_type, 'bbox')
        self.assertIsInstance(module.train_set.dataset, kraken.lib.dataset.GroundTruthDataset)
        trainer = KrakenTrainer(max_steps=1)
        self.assertEqual(module.nn.named_spec[-1].split("c")[-1], '16')

    def test_krakentrainer_rec_box_append(self):
        """
        Tests that appending new layers onto a loaded model works.
        """
        training_data = self.box_lines
        evaluation_data = self.box_lines
        module = RecognitionModel(format_type='path',
                                  model=self.model,
                                  append=1,
                                  spec='[Cr4,4,32]',
                                  training_data=training_data,
                                  evaluation_data=evaluation_data)
        module.setup()
        self.assertEqual(module.nn.seg_type, 'bbox')
        self.assertIsInstance(module.train_set.dataset, kraken.lib.dataset.GroundTruthDataset)
        self.assertTrue(module.nn.spec.startswith('[1,48,0,1 Cr{C_0}4,2,1,4,2 Cr{C_1}4,4,32 O{O_2}'))
        trainer = KrakenTrainer(max_steps=1)

    def test_krakentrainer_rec_bl_load(self):
        training_data = [self.xml]
        evaluation_data = [self.xml]
        module = RecognitionModel(format_type='xml',
                                  model=self.model,
                                  training_data=training_data,
                                  evaluation_data=evaluation_data,
                                  resize='fail')
        with raises(KrakenInputException):
            module.setup()

    def test_krakentrainer_rec_bl_load_union(self):
        training_data = [self.xml]
        evaluation_data = [self.xml]
        module = RecognitionModel(format_type='xml',
                                  model=self.model,
                                  training_data=training_data,
                                  evaluation_data=evaluation_data,
                                  resize='union')
        module.setup()
        self.assertEqual(module.nn.seg_type, 'baselines')
        self.assertIsInstance(module.train_set.dataset, kraken.lib.dataset.PolygonGTDataset)
        trainer = KrakenTrainer(max_steps=1)
        self.assertEqual(module.nn.named_spec[-1].split("c")[-1], '60')

    def test_krakentrainer_rec_bl_load_new(self):
        training_data = [self.xml]
        evaluation_data = [self.xml]
        module = RecognitionModel(format_type='xml',
                                  model=self.model,
                                  training_data=training_data,
                                  evaluation_data=evaluation_data,
                                  resize='new')
        module.setup()
        self.assertEqual(module.nn.seg_type, 'baselines')
        self.assertIsInstance(module.train_set.dataset, kraken.lib.dataset.PolygonGTDataset)
        trainer = KrakenTrainer(max_steps=1)
        self.assertEqual(module.nn.named_spec[-1].split("c")[-1], '60')

    def test_krakentrainer_rec_bl_append(self):
        training_data = [self.xml]
        evaluation_data = [self.xml]
        module = RecognitionModel(format_type='xml',
                                  model=self.model,
                                  append=1,
                                  spec='[Cr4,4,32]',
                                  training_data=training_data,
                                  evaluation_data=evaluation_data)
        module.setup()
        self.assertEqual(module.nn.seg_type, 'baselines')
        self.assertIsInstance(module.train_set.dataset, kraken.lib.dataset.PolygonGTDataset)
        self.assertTrue(module.nn.spec.startswith('[1,48,0,1 Cr{C_0}4,2,1,4,2 Cr{C_1}4,4,32 O{O_2}'))
        trainer = KrakenTrainer(max_steps=1)

    def test_krakentrainer_rec_box_path(self):
        """
        Tests recognition trainer constructor with legacy path training data.
        """
        training_data = self.box_lines
        evaluation_data = self.box_lines
        module = RecognitionModel(format_type='path',
                                  training_data=training_data,
                                  evaluation_data=evaluation_data)
        module.setup()
        self.assertEqual(module.nn.seg_type, 'bbox')
        self.assertIsInstance(module.train_set.dataset, kraken.lib.dataset.GroundTruthDataset)
        trainer = KrakenTrainer(max_steps=1)

    def test_krakentrainer_rec_bl_xml(self):
        """
        Tests recognition trainer constructor with XML training data.
        """
        training_data = [self.xml]
        evaluation_data = [self.xml]
        module = RecognitionModel(format_type='xml',
                                  training_data=training_data,
                                  evaluation_data=evaluation_data)
        module.setup()
        self.assertEqual(module.nn.seg_type, 'baselines')
        self.assertIsInstance(module.train_set.dataset, kraken.lib.dataset.PolygonGTDataset)
        self.assertEqual(len(module.train_set.dataset), 44)
        self.assertEqual(len(module.val_set.dataset), 44)
        trainer = KrakenTrainer(max_steps=1)

    def test_krakentrainer_rec_bl_dict(self):
        """
        Tests recognition trainer constructor with dictionary style training data.
        """
        training_data = [{'image': resources / 'bw.png', 'text': 'foo', 'baseline': [[10, 10], [300, 10]], 'boundary': [[10, 5], [300, 5], [300, 15], [10, 15]]}]
        evaluation_data = [{'image': resources / 'bw.png', 'text': 'foo', 'baseline': [[10, 10], [300, 10]], 'boundary': [[10, 5], [300, 5], [300, 15], [10, 15]]}]
        module = RecognitionModel(format_type=None,
                                  training_data=training_data,
                                  evaluation_data=evaluation_data)
        module.setup()
        self.assertEqual(module.nn.seg_type, 'baselines')
        self.assertIsInstance(module.train_set.dataset, kraken.lib.dataset.PolygonGTDataset)
        trainer = KrakenTrainer(max_steps=1)

    def test_krakentrainer_rec_bl_augment(self):
        """
        Test that augmentation is added if specified.
        """
        training_data = [self.xml]
        evaluation_data = [self.xml]
        module = RecognitionModel(format_type='xml',
                                  training_data=training_data,
                                  evaluation_data=evaluation_data)
        module.setup()
        self.assertEqual(module.train_set.dataset.aug, None)

        module = RecognitionModel({'augment': True},
                                  format_type='xml',
                                  training_data=training_data,
                                  evaluation_data=evaluation_data)
        module.setup()
        self.assertIsInstance(module.train_set.dataset.aug, kraken.lib.dataset.recognition.DefaultAugmenter)

    def test_krakentrainer_rec_box_augment(self):
        """
        Test that augmentation is added if specified.
        """
        training_data = self.box_lines
        evaluation_data = self.box_lines
        module = RecognitionModel(format_type='path',
                                  training_data=training_data,
                                  evaluation_data=evaluation_data)
        module.setup()
        self.assertEqual(module.train_set.dataset.aug, None)

        module = RecognitionModel({'augment': True},
                                  format_type='path',
                                  training_data=training_data,
                                  evaluation_data=evaluation_data)
        module.setup()
        self.assertIsInstance(module.train_set.dataset.aug, kraken.lib.dataset.recognition.DefaultAugmenter)



#
# Copyright 2015 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
Training loop interception helpers
"""
import logging
import re
import warnings
from typing import (TYPE_CHECKING, Any, Callable, Dict, Literal, Optional,
                    Sequence, Union)
from functools import partial

import numpy as np
import lightning as L
import torch
import torch.nn.functional as F
from lightning.pytorch.callbacks import (BaseFinetuning, Callback,
                                         EarlyStopping, LearningRateMonitor)
from torch.optim import lr_scheduler
from torch.utils.data import DataLoader, Subset, random_split
from torchmetrics.classification import (MultilabelAccuracy,
                                         MultilabelJaccardIndex)
from torchmetrics.text import CharErrorRate, WordErrorRate

from kraken.containers import Segmentation
from kraken.lib import default_specs, models, progress, vgsl
from kraken.lib.codec import PytorchCodec
from kraken.lib.dataset import (ArrowIPCRecognitionDataset, BaselineSet,
                                GroundTruthDataset, ImageInputTransforms,
                                PolygonGTDataset, collate_sequences)
from kraken.lib.exceptions import KrakenEncodeException, KrakenInputException
from kraken.lib.models import validate_hyper_parameters
from kraken.lib.util import make_printable, parse_gt_path
from kraken.lib.xml import XMLPage

if TYPE_CHECKING:
    from os import PathLike

logger = logging.getLogger(__name__)


def _star_fun(fun, kwargs):
    try:
        return fun(**kwargs)
    except FileNotFoundError as e:
        logger.warning(f'{e.strerror}: {e.filename}. Skipping.')
    except KrakenInputException as e:
        logger.warning(str(e))
    return None


def _validation_worker_init_fn(worker_id):
    """ Fix random seeds so that augmentation always produces the same
        results when validating. Temporarily increase the logging level
        for lightning because otherwise it will display a message
        at info level about the seed being changed. """
    from lightning.pytorch import seed_everything
    seed_everything(42)


class KrakenTrainer(L.Trainer):
    def __init__(self,
                 enable_progress_bar: bool = True,
                 enable_summary: bool = True,
                 min_epochs: int = 5,
                 max_epochs: int = 100,
                 freeze_backbone=-1,
                 pl_logger: Union[L.pytorch.loggers.logger.Logger, str, None] = None,
                 log_dir: Optional['PathLike'] = None,
                 *args,
                 **kwargs):
        kwargs['enable_checkpointing'] = False
        kwargs['enable_progress_bar'] = enable_progress_bar
        kwargs['min_epochs'] = min_epochs
        kwargs['max_epochs'] = max_epochs
        kwargs['callbacks'] = ([] if 'callbacks' not in kwargs else kwargs['callbacks'])
        if not isinstance(kwargs['callbacks'], list):
            kwargs['callbacks'] = [kwargs['callbacks']]

        if pl_logger:
            if 'logger' in kwargs and isinstance(kwargs['logger'], L.pytorch.loggers.logger.Logger):
                logger.debug('Experiment logger has been provided outside KrakenTrainer as `logger`')
            elif isinstance(pl_logger, L.pytorch.loggers.logger.Logger):
                logger.debug('Experiment logger has been provided outside KrakenTrainer as `pl_logger`')
                kwargs['logger'] = pl_logger
            elif pl_logger == 'tensorboard':
                logger.debug('Creating default experiment logger')
                kwargs['logger'] = L.pytorch.loggers.TensorBoardLogger(log_dir)
            else:
                logger.error('`pl_logger` was set, but %s is not an accepted value', pl_logger)
                raise ValueError(f'{pl_logger} is not acceptable as logger')
            kwargs['callbacks'].append(LearningRateMonitor(logging_interval='step'))
        else:
            kwargs['logger'] = False

        if enable_progress_bar:
            progress_bar_cb = progress.KrakenTrainProgressBar(leave=True)
            kwargs['callbacks'].append(progress_bar_cb)

        if enable_summary:
            from lightning.pytorch.callbacks import RichModelSummary
            summary_cb = RichModelSummary(max_depth=2)
            kwargs['callbacks'].append(summary_cb)
            kwargs['enable_model_summary'] = False

        if freeze_backbone > 0:
            kwargs['callbacks'].append(KrakenFreezeBackbone(freeze_backbone))

        kwargs['callbacks'].extend([KrakenSetOneChannelMode(), KrakenSaveModel()])
        super().__init__(*args, **kwargs)
        self.automatic_optimization = False

    def fit(self, *args, **kwargs):
        with warnings.catch_warnings():
            warnings.filterwarnings(action='ignore', category=UserWarning,
                                    message='The dataloader,')
            super().fit(*args, **kwargs)


class KrakenFreezeBackbone(BaseFinetuning):
    """
    Callback freezing all but the last layer for fixed number of iterations.
    """
    def __init__(self, unfreeze_at_iterations=10):
        super().__init__()
        self.unfreeze_at_iteration = unfreeze_at_iterations

    def freeze_before_training(self, pl_module):
        pass

    def finetune_function(self, pl_module, current_epoch, optimizer):
        pass

    def on_train_start(self, trainer: "L.Trainer", pl_module: "L.LightningModule") -> None:
        self.freeze(pl_module.net[:-1])

    def on_train_batch_start(self, trainer: "L.Trainer", pl_module: "L.LightningModule", batch, batch_idx) -> None:
        """
        Called for each training batch.
        """
        if trainer.global_step == self.unfreeze_at_iteration:
            for opt_idx, optimizer in enumerate(trainer.optimizers):
                num_param_groups = len(optimizer.param_groups)
                self.unfreeze_and_add_param_group(modules=pl_module.net[:-1],
                                                  optimizer=optimizer,
                                                  train_bn=True,)
                current_param_groups = optimizer.param_groups
                self._store(pl_module, opt_idx, num_param_groups, current_param_groups)

    def on_train_epoch_start(self, trainer: "L.Trainer", pl_module: "L.LightningModule") -> None:
        """Called when the epoch begins."""
        pass


class KrakenSetOneChannelMode(Callback):
    """
    Callback that sets the one_channel_mode of the model after the first epoch.
    """
    def on_train_epoch_end(self, trainer: "L.Trainer", pl_module: "L.LightningModule") -> None:
        # fill one_channel_mode after 1 iteration over training data set
        if not trainer.sanity_checking and trainer.current_epoch == 0 and trainer.model.nn.model_type == 'recognition':
            ds = getattr(pl_module, 'train_set', None)
            if not ds and trainer.datamodule:
                ds = trainer.datamodule.train_set
            im_mode = ds.dataset.im_mode
            if im_mode in ['1', 'L']:
                logger.info(f'Setting model one_channel_mode to {im_mode}.')
                trainer.model.nn.one_channel_mode = im_mode


class KrakenSaveModel(Callback):
    """
    Kraken's own serialization callback instead of pytorch's.
    """
    def on_validation_end(self, trainer: "L.Trainer", pl_module: "L.LightningModule") -> None:
        if not trainer.sanity_checking:
            trainer.model.nn.hyper_params['completed_epochs'] += 1
            metric = float(trainer.logged_metrics['val_metric']) if 'val_metric' in trainer.logged_metrics else -1.0
            trainer.model.nn.user_metadata['accuracy'].append((trainer.global_step, metric))
            trainer.model.nn.user_metadata['metrics'].append((trainer.global_step, {k: float(v) for k, v in trainer.logged_metrics.items()}))

            logger.info('Saving to {}_{}.mlmodel'.format(trainer.model.output, trainer.current_epoch))
            trainer.model.nn.save_model(f'{trainer.model.output}_{trainer.current_epoch}.mlmodel')
            trainer.model.best_model = f'{trainer.model.output}_{trainer.model.best_epoch}.mlmodel'


class RecognitionModel(L.LightningModule):
    def __init__(self,
                 hyper_params: Dict[str, Any] = None,
                 output: str = 'model',
                 spec: str = default_specs.RECOGNITION_SPEC,
                 append: Optional[int] = None,
                 model: Optional[Union['PathLike', str]] = None,
                 reorder: Union[bool, str] = True,
                 training_data: Union[Sequence[Union['PathLike', str]], Sequence[Dict[str, Any]]] = None,
                 evaluation_data: Optional[Union[Sequence[Union['PathLike', str]], Sequence[Dict[str, Any]]]] = None,
                 partition: Optional[float] = 0.9,
                 binary_dataset_split: bool = False,
                 num_workers: int = 1,
                 load_hyper_parameters: bool = False,
                 force_binarization: bool = False,
                 format_type: Literal['path', 'alto', 'page', 'xml', 'binary'] = 'path',
                 codec: Optional[Dict] = None,
                 resize: Literal['fail', 'both', 'new', 'add', 'union'] = 'fail',
                 legacy_polygons: bool = False):
        """
        A LightningModule encapsulating the training setup for a text
        recognition model.

        Setup parameters (load, training_data, evaluation_data, ....) are
        named, model hyperparameters (everything in
        `kraken.lib.default_specs.RECOGNITION_HYPER_PARAMS`) are in in the
        `hyper_params` argument.

        Args:
            hyper_params (dict): Hyperparameter dictionary containing all fields
                                 from
                                 kraken.lib.default_specs.RECOGNITION_HYPER_PARAMS
            **kwargs: Setup parameters, i.e. CLI parameters of the train() command.
        """
        super().__init__()
        self.legacy_polygons = legacy_polygons
        hyper_params_ = default_specs.RECOGNITION_HYPER_PARAMS.copy()
        if model:
            logger.info(f'Loading existing model from {model} ')
            self.nn = vgsl.TorchVGSLModel.load_model(model)

            if self.nn.model_type not in [None, 'recognition']:
                raise ValueError(f'Model {model} is of type {self.nn.model_type} while `recognition` is expected.')

            if load_hyper_parameters:
                hp = self.nn.hyper_params
            else:
                hp = {}
            hyper_params_.update(hp)
        else:
            self.nn = None

        if hyper_params:
            hyper_params_.update(hyper_params)
        self.hyper_params = hyper_params_
        self.save_hyperparameters()

        self.reorder = reorder
        self.append = append
        self.model = model
        self.num_workers = num_workers
        if resize == "add":
            resize = "union"
            warnings.warn("'add' value for resize has been deprecated. Use 'union' instead.", DeprecationWarning)
        elif resize == "both":
            resize = "new"
            warnings.warn("'both' value for resize has been deprecated. Use 'new' instead.", DeprecationWarning)

        self.resize = resize
        self.format_type = format_type
        self.output = output

        self.best_epoch = -1
        self.best_metric = 0.0
        self.best_model = None

        DatasetClass = GroundTruthDataset
        valid_norm = True
        if format_type in ['xml', 'page', 'alto']:
            logger.info(f'Parsing {len(training_data)} XML files for training data')
            training_data = [{'page': XMLPage(file, format_type).to_container()} for file in training_data]
            if evaluation_data:
                logger.info(f'Parsing {len(evaluation_data)} XML files for validation data')
                evaluation_data = [{'page': XMLPage(file, format_type).to_container()} for file in evaluation_data]
            if binary_dataset_split:
                logger.warning('Internal binary dataset splits are enabled but using non-binary dataset files. Will be ignored.')
                binary_dataset_split = False
            DatasetClass = partial(PolygonGTDataset, legacy_polygons=legacy_polygons)
            valid_norm = False
        elif format_type == 'binary':
            DatasetClass = ArrowIPCRecognitionDataset
            valid_norm = False
            logger.info(f'Got {len(training_data)} binary dataset files for training data')
            training_data = [{'file': file} for file in training_data]
            if evaluation_data:
                logger.info(f'Got {len(evaluation_data)} binary dataset files for validation data')
                evaluation_data = [{'file': file} for file in evaluation_data]
        elif format_type == 'path':
            if force_binarization:
                logger.warning('Forced binarization enabled in `path` mode. Will be ignored.')
                force_binarization = False
            if binary_dataset_split:
                logger.warning('Internal binary dataset splits are enabled but using non-binary dataset files. Will be ignored.')
                binary_dataset_split = False
            logger.info(f'Got {len(training_data)} line strip images for training data')
            training_data = [{'line': parse_gt_path(im)} for im in training_data]
            if evaluation_data:
                logger.info(f'Got {len(evaluation_data)} line strip images for validation data')
                evaluation_data = [{'line': parse_gt_path(im)} for im in evaluation_data]
            valid_norm = True
        # format_type is None. Determine training type from container class types
        elif not format_type:
            if training_data[0].type == 'baselines':
                DatasetClass = partial(PolygonGTDataset, legacy_polygons=legacy_polygons)
                valid_norm = False
            else:
                if force_binarization:
                    logger.warning('Forced binarization enabled with box lines. Will be ignored.')
                    force_binarization = False
                if binary_dataset_split:
                    logger.warning('Internal binary dataset splits are enabled but using non-binary dataset files. Will be ignored.')
                    binary_dataset_split = False
            samples = []
            for sample in training_data:
                if isinstance(sample, Segmentation):
                    samples.append({'page': sample})
                else:
                    samples.append({'line': sample})
            training_data = samples
            if evaluation_data:
                samples = []
                for sample in evaluation_data:
                    if isinstance(sample, Segmentation):
                        samples.append({'page': sample})
                    else:
                        samples.append({'line': sample})
                evaluation_data = samples
        else:
            raise ValueError(f'format_type {format_type} not in [alto, page, xml, path, binary].')

        spec = spec.strip()
        if spec[0] != '[' or spec[-1] != ']':
            raise ValueError(f'VGSL spec {spec} not bracketed')
        self.spec = spec
        # preparse input sizes from vgsl string to seed ground truth data set
        # sizes and dimension ordering.
        if not self.nn:
            blocks = spec[1:-1].split(' ')
            m = re.match(r'(\d+),(\d+),(\d+),(\d+)', blocks[0])
            if not m:
                raise ValueError(f'Invalid input spec {blocks[0]}')
            batch, height, width, channels = [int(x) for x in m.groups()]
        else:
            batch, channels, height, width = self.nn.input

        self.transforms = ImageInputTransforms(batch,
                                               height,
                                               width,
                                               channels,
                                               (self.hparams.hyper_params['pad'], 0),
                                               valid_norm,
                                               force_binarization)

        self.example_input_array = torch.Tensor(batch,
                                                channels,
                                                height if height else 32,
                                                width if width else 400)

        if 'file_system' in torch.multiprocessing.get_all_sharing_strategies():
            logger.debug('Setting multiprocessing tensor sharing strategy to file_system')
            torch.multiprocessing.set_sharing_strategy('file_system')

        val_set = None
        if evaluation_data:
            train_set = self._build_dataset(DatasetClass, training_data)
            self.train_set = Subset(train_set, range(len(train_set)))
            val_set = self._build_dataset(DatasetClass, evaluation_data)
            self.val_set = Subset(val_set, range(len(val_set)))
        elif binary_dataset_split:
            train_set = self._build_dataset(DatasetClass, training_data, split_filter='train')
            self.train_set = Subset(train_set, range(len(train_set)))
            val_set = self._build_dataset(DatasetClass, training_data, split_filter='validation')
            self.val_set = Subset(val_set, range(len(val_set)))
            logger.info(f'Found {len(self.train_set)} (train) / {len(self.val_set)} (val) samples in pre-encoded dataset')
        else:
            train_set = self._build_dataset(DatasetClass, training_data)
            train_len = int(len(train_set)*partition)
            val_len = len(train_set) - train_len
            logger.info(f'No explicit validation data provided. Splitting off '
                        f'{val_len} (of {len(train_set)}) samples to validation '
                        'set. (Will disable alphabet mismatch detection.)')
            self.train_set, self.val_set = random_split(train_set, (train_len, val_len))

        if len(self.train_set) == 0 or len(self.val_set) == 0:
            raise ValueError('No valid training data was provided to the train '
                             'command. Please add valid XML, line, or binary data.')

        if format_type == 'binary':
            legacy_train_status = self.train_set.dataset.legacy_polygons_status
            if self.val_set.dataset.legacy_polygons_status != legacy_train_status:
                logger.warning('Train and validation set have different legacy '
                               f'polygon status: {legacy_train_status} and '
                               f'{self.val_set.dataset.legacy_polygons_status}. Train set '
                               'status prevails.')
            if legacy_train_status == "mixed":
                logger.warning('Mixed legacy polygon status in training dataset. Consider recompilation.')
                legacy_train_status = False
            if legacy_polygons != legacy_train_status:
                logger.warning(f'Setting dataset legacy polygon status to {legacy_train_status} based on training set.')
                self.legacy_polygons = legacy_train_status

        logger.info(f'Training set {len(self.train_set)} lines, validation set '
                    f'{len(self.val_set)} lines, alphabet {len(train_set.alphabet)} '
                    'symbols')
        alpha_diff_only_train = set(self.train_set.dataset.alphabet).difference(set(self.val_set.dataset.alphabet))
        alpha_diff_only_val = set(self.val_set.dataset.alphabet).difference(set(self.train_set.dataset.alphabet))
        if alpha_diff_only_train:
            logger.warning(f'alphabet mismatch: chars in training set only: '
                           f'{alpha_diff_only_train} (not included in accuracy test '
                           'during training)')
        if alpha_diff_only_val:
            logger.warning(f'alphabet mismatch: chars in validation set only: {alpha_diff_only_val} (not trained)')
        logger.info('grapheme\tcount')
        for k, v in sorted(train_set.alphabet.items(), key=lambda x: x[1], reverse=True):
            char = make_printable(k)
            if char == k:
                char = '\t' + char
            logger.info(f'{char}\t{v}')

        if codec:
            logger.info('Instantiating codec')
            self.codec = PytorchCodec(codec)
            for k, v in self.codec.c2l.items():
                char = make_printable(k)
                if char == k:
                    char = '\t' + char
                logger.info(f'{char}\t{v}')
        else:
            self.codec = None

        logger.info('Encoding training set')

        self.val_cer = CharErrorRate()
        self.val_wer = WordErrorRate()

    def _build_dataset(self,
                       DatasetClass,
                       training_data,
                       **kwargs):
        dataset = DatasetClass(normalization=self.hparams.hyper_params['normalization'],
                               whitespace_normalization=self.hparams.hyper_params['normalize_whitespace'],
                               reorder=self.reorder,
                               im_transforms=self.transforms,
                               augmentation=self.hparams.hyper_params['augment'],
                               **kwargs)

        for sample in training_data:
            try:
                dataset.add(**sample)
            except KrakenInputException as e:
                logger.warning(str(e))
        if self.format_type == 'binary' and (self.hparams.hyper_params['normalization'] or
                                             self.hparams.hyper_params['normalize_whitespace'] or
                                             self.reorder):
            logger.debug('Text transformations modifying alphabet selected. Rebuilding alphabet')
            dataset.rebuild_alphabet()

        return dataset

    def forward(self, x, seq_lens=None):
        return self.net(x, seq_lens)

    def training_step(self, batch, batch_idx):
        input, target = batch['image'], batch['target']
        # sequence batch
        if 'seq_lens' in batch:
            seq_lens, label_lens = batch['seq_lens'], batch['target_lens']
            target = (target, label_lens)
            o = self.net(input, seq_lens)
        else:
            o = self.net(input)

        seq_lens = o[1]
        output = o[0]
        target_lens = target[1]
        target = target[0]
        # height should be 1 by now
        if output.size(2) != 1:
            raise KrakenInputException('Expected dimension 3 to be 1, actual {}'.format(output.size(2)))
        output = output.squeeze(2)
        # NCW -> WNC
        loss = self.nn.criterion(output.permute(2, 0, 1),  # type: ignore
                                 target,
                                 seq_lens,
                                 target_lens)
        self.log('train_loss', loss, on_step=True, on_epoch=False, prog_bar=False, logger=True)
        return loss

    def validation_step(self, batch, batch_idx):
        pred = self.rec_nn.predict_string(batch['image'], batch['seq_lens'])
        idx = 0
        decoded_targets = []
        for offset in batch['target_lens']:
            decoded_targets.append(''.join([x[0] for x in self.val_codec.decode([(x, 0, 0, 0) for x in batch['target'][idx:idx+offset]])]))
            idx += offset
        self.val_cer.update(pred, decoded_targets)
        self.val_wer.update(pred, decoded_targets)

        if self.logger and self.trainer.state.stage != 'sanity_check' and self.hparams.hyper_params["batch_size"] * batch_idx < 16:
            for i in range(self.hparams.hyper_params["batch_size"]):
                count = self.hparams.hyper_params["batch_size"] * batch_idx + i
                if count < 16:
                    self.logger.experiment.add_image(f'Validation #{count}, target: {decoded_targets[i]}',
                                                     batch['image'][i],
                                                     self.global_step,
                                                     dataformats="CHW")
                    self.logger.experiment.add_text(f'Validation #{count}, target: {decoded_targets[i]}',
                                                    pred[i],
                                                    self.global_step)

    def on_validation_epoch_end(self):
        if not self.trainer.sanity_checking:
            accuracy = 1.0 - self.val_cer.compute()
            word_accuracy = 1.0 - self.val_wer.compute()

            if accuracy > self.best_metric:
                logger.debug(f'Updating best metric from {self.best_metric} ({self.best_epoch}) to {accuracy} ({self.current_epoch})')
                self.best_epoch = self.current_epoch
                self.best_metric = accuracy
            logger.info(f'validation run: total chars {self.val_cer.total} errors {self.val_cer.errors} accuracy {accuracy}')
            self.log('val_accuracy', accuracy, on_step=False, on_epoch=True, prog_bar=True, logger=True)
            self.log('val_word_accuracy', word_accuracy, on_step=False, on_epoch=True, prog_bar=True, logger=True)
            self.log('val_metric', accuracy, on_step=False, on_epoch=True, prog_bar=False, logger=True)
        # reset metrics even if not sanity checking
        self.val_cer.reset()
        self.val_wer.reset()

    def setup(self, stage: Optional[str] = None):
        # finalize models in case of appending/loading
        if stage in [None, 'fit']:

            # Log a few sample images before the datasets are encoded.
            # This is only possible for Arrow datasets, because the
            # other dataset types can only be accessed after encoding
            if self.logger and isinstance(self.train_set.dataset, ArrowIPCRecognitionDataset):
                for i in range(min(len(self.train_set), 16)):
                    idx = np.random.randint(len(self.train_set))
                    sample = self.train_set[idx]
                    self.logger.experiment.add_image(f'train_set sample #{i}: {sample["target"]}', sample['image'])

            if self.append:
                self.train_set.dataset.encode(self.codec)
                # now we can create a new model
                self.spec = '[{} O1c{}]'.format(self.spec[1:-1], self.train_set.dataset.codec.max_label + 1)
                logger.info(f'Appending {self.spec} to existing model {self.nn.spec} after {self.append}')
                self.nn.append(self.append, self.spec)
                self.nn.add_codec(self.train_set.dataset.codec)
                logger.info(f'Assembled model spec: {self.nn.spec}')
            elif self.model:
                self.spec = self.nn.spec

                # prefer explicitly given codec over network codec if mode is 'new'
                codec = self.codec if (self.codec and self.resize == 'new') else self.nn.codec

                codec.strict = True

                try:
                    self.train_set.dataset.encode(codec)
                except KrakenEncodeException:
                    alpha_diff = set(self.train_set.dataset.alphabet).difference(
                        set(codec.c2l.keys())
                    )
                    if self.resize == 'fail':
                        raise KrakenInputException(f'Training data and model codec alphabets mismatch: {alpha_diff}')
                    elif self.resize == 'union':
                        logger.info(f'Resizing codec to include '
                                    f'{len(alpha_diff)} new code points')
                        # Construct two codecs:
                        # 1. training codec containing only the vocabulary in the training dataset
                        # 2. validation codec = training codec + validation set vocabulary
                        # This keep the codec in the model from being 'polluted' by non-trained characters.
                        train_codec = codec.add_labels(alpha_diff)
                        self.nn.add_codec(train_codec)
                        logger.info(f'Resizing last layer in network to {train_codec.max_label+1} outputs')
                        self.nn.resize_output(train_codec.max_label + 1)
                        self.train_set.dataset.encode(train_codec)
                    elif self.resize == 'new':
                        logger.info(f'Resizing network or given codec to '
                                    f'{len(self.train_set.dataset.alphabet)} '
                                    f'code sequences')
                        # same codec procedure as above, just with merging.
                        self.train_set.dataset.encode(None)
                        train_codec, del_labels = codec.merge(self.train_set.dataset.codec)
                        # Switch codec.
                        self.nn.add_codec(train_codec)
                        logger.info(f'Deleting {len(del_labels)} output classes from network '
                                    f'({len(codec)-len(del_labels)} retained)')
                        self.nn.resize_output(train_codec.max_label + 1, del_labels)
                        self.train_set.dataset.encode(train_codec)
                    else:
                        raise ValueError(f'invalid resize parameter value {self.resize}')
                self.nn.codec.strict = False
                self.spec = self.nn.spec
            else:
                self.train_set.dataset.encode(self.codec)
                logger.info(f'Creating new model {self.spec} with {self.train_set.dataset.codec.max_label+1} outputs')
                self.spec = '[{} O1c{}]'.format(self.spec[1:-1], self.train_set.dataset.codec.max_label + 1)
                self.nn = vgsl.TorchVGSLModel(self.spec)
                self.nn.use_legacy_polygons = self.legacy_polygons
                # initialize weights
                self.nn.init_weights()
                self.nn.add_codec(self.train_set.dataset.codec)

            val_diff = set(self.val_set.dataset.alphabet).difference(
                set(self.train_set.dataset.codec.c2l.keys())
            )
            logger.info(f'Adding {len(val_diff)} dummy labels to validation set codec.')

            val_codec = self.nn.codec.add_labels(val_diff)
            self.val_set.dataset.encode(val_codec)
            self.val_codec = val_codec

            if self.nn.one_channel_mode and self.train_set.dataset.im_mode != self.nn.one_channel_mode:
                logger.warning(f'Neural network has been trained on mode {self.nn.one_channel_mode} images, '
                               f'training set contains mode {self.train_set.dataset.im_mode} data. Consider setting `force_binarization`')

            if self.format_type != 'path' and self.nn.seg_type == 'bbox':
                logger.warning('Neural network has been trained on bounding box image information but training set is polygonal.')

            self.nn.hyper_params = self.hparams.hyper_params
            self.nn.model_type = 'recognition'

            if not self.nn.seg_type:
                logger.info(f'Setting seg_type to {self.train_set.dataset.seg_type}.')
                self.nn.seg_type = self.train_set.dataset.seg_type

            self.rec_nn = models.TorchSeqRecognizer(self.nn, train=None, device=None)
            self.net = self.nn.nn

            torch.set_num_threads(max(self.num_workers, 1))

    def train_dataloader(self):
        return DataLoader(self.train_set,
                          batch_size=self.hparams.hyper_params['batch_size'],
                          num_workers=self.num_workers,
                          pin_memory=True,
                          shuffle=True,
                          collate_fn=collate_sequences)

    def val_dataloader(self):
        return DataLoader(self.val_set,
                          shuffle=False,
                          batch_size=self.hparams.hyper_params['batch_size'],
                          num_workers=self.num_workers,
                          pin_memory=True,
                          collate_fn=collate_sequences,
                          worker_init_fn=_validation_worker_init_fn)

    def configure_callbacks(self):
        callbacks = []
        if self.hparams.hyper_params['quit'] == 'early':
            callbacks.append(EarlyStopping(monitor='val_accuracy',
                                           mode='max',
                                           patience=self.hparams.hyper_params['lag'],
                                           stopping_threshold=1.0))

        return callbacks

    # configuration of optimizers and learning rate schedulers
    # --------------------------------------------------------
    #
    # All schedulers are created internally with a frequency of step to enable
    # batch-wise learning rate warmup. In lr_scheduler_step() calls to the
    # scheduler are then only performed at the end of the epoch.
    def configure_optimizers(self):
        return _configure_optimizer_and_lr_scheduler(self.hparams.hyper_params,
                                                     self.nn.nn.parameters(),
                                                     len_train_set=len(self.train_set),
                                                     loss_tracking_mode='max')

    def optimizer_step(self, epoch, batch_idx, optimizer, optimizer_closure):
        # update params
        optimizer.step(closure=optimizer_closure)

        # linear warmup between 0 and the initial learning rate `lrate` in `warmup`
        # steps.
        if self.hparams.hyper_params['warmup'] and self.trainer.global_step < self.hparams.hyper_params['warmup']:
            lr_scale = min(1.0, float(self.trainer.global_step + 1) / self.hparams.hyper_params['warmup'])
            for pg in optimizer.param_groups:
                pg["lr"] = lr_scale * self.hparams.hyper_params['lrate']

    def lr_scheduler_step(self, scheduler, metric):
        if not self.hparams.hyper_params['warmup'] or self.trainer.global_step >= self.hparams.hyper_params['warmup']:
            # step OneCycleLR each batch if not in warmup phase
            if isinstance(scheduler, lr_scheduler.OneCycleLR):
                scheduler.step()
            # step every other scheduler epoch-wise
            elif self.trainer.is_last_batch:
                if metric is None:
                    scheduler.step()
                else:
                    scheduler.step(metric)


class SegmentationModel(L.LightningModule):
    def __init__(self,
                 hyper_params: Dict = None,
                 load_hyper_parameters: bool = False,
                 progress_callback: Callable[[str, int], Callable[[None], None]] = lambda string, length: lambda: None,
                 message: Callable[[str], None] = lambda *args, **kwargs: None,
                 output: str = 'model',
                 spec: str = default_specs.SEGMENTATION_SPEC,
                 model: Optional[Union['PathLike', str]] = None,
                 training_data: Union[Sequence[Union['PathLike', str]], Sequence[Segmentation]] = None,
                 evaluation_data: Optional[Union[Sequence[Union['PathLike', str]], Sequence[Segmentation]]] = None,
                 partition: Optional[float] = 0.9,
                 num_workers: int = 1,
                 force_binarization: bool = False,
                 format_type: Literal['path', 'alto', 'page', 'xml', None] = 'path',
                 suppress_regions: bool = False,
                 suppress_baselines: bool = False,
                 valid_regions: Optional[Sequence[str]] = None,
                 valid_baselines: Optional[Sequence[str]] = None,
                 merge_regions: Optional[Dict[str, str]] = None,
                 merge_baselines: Optional[Dict[str, str]] = None,
                 bounding_regions: Optional[Sequence[str]] = None,
                 resize: Literal['fail', 'both', 'new', 'add', 'union'] = 'fail',
                 topline: Union[bool, None] = False):
        """
        A LightningModule encapsulating the training setup for a page
        segmentation model.

        Setup parameters (load, training_data, evaluation_data, ....) are
        named, model hyperparameters (everything in
        `kraken.lib.default_specs.SEGMENTATION_HYPER_PARAMS`) are in in the
        `hyper_params` argument.

        Args:
            hyper_params (dict): Hyperparameter dictionary containing all fields
                                 from
                                 kraken.lib.default_specs.SEGMENTATION_HYPER_PARAMS
            **kwargs: Setup parameters, i.e. CLI parameters of the segtrain() command.
        """

        super().__init__()

        self.best_epoch = -1
        self.best_metric = 0.0
        self.best_model = None

        self.model = model
        self.num_workers = num_workers

        if resize == "add":
            resize = "union"
            warnings.warn("'add' value for resize has been deprecated. Use 'union' instead.", DeprecationWarning)
        elif resize == "both":
            resize = "new"
            warnings.warn("'both' value for resize has been deprecated. Use 'new' instead.", DeprecationWarning)
        self.resize = resize

        self.output = output
        self.bounding_regions = bounding_regions
        self.topline = topline

        hyper_params_ = default_specs.SEGMENTATION_HYPER_PARAMS.copy()

        if model:
            logger.info(f'Loading existing model from {model}')
            self.nn = vgsl.TorchVGSLModel.load_model(model)

            if self.nn.model_type not in [None, 'segmentation']:
                raise ValueError(f'Model {model} is of type {self.nn.model_type} while `segmentation` is expected.')

            if load_hyper_parameters:
                hp = self.nn.hyper_params
            else:
                hp = {}
            hyper_params_.update(hp)
            batch, channels, height, width = self.nn.input
        else:
            self.nn = None

            spec = spec.strip()
            if spec[0] != '[' or spec[-1] != ']':
                raise ValueError(f'VGSL spec "{spec}" not bracketed')
            self.spec = spec
            blocks = spec[1:-1].split(' ')
            m = re.match(r'(\d+),(\d+),(\d+),(\d+)', blocks[0])
            if not m:
                raise ValueError(f'Invalid input spec {blocks[0]}')
            batch, height, width, channels = [int(x) for x in m.groups()]

        if hyper_params:
            hyper_params_.update(hyper_params)

        validate_hyper_parameters(hyper_params_)
        self.hyper_params = hyper_params_
        self.save_hyperparameters()

        if format_type in ['xml', 'page', 'alto']:
            logger.info(f'Parsing {len(training_data)} XML files for training data')
            training_data = [XMLPage(file, format_type).to_container() for file in training_data]
            if evaluation_data:
                logger.info(f'Parsing {len(evaluation_data)} XML files for validation data')
                evaluation_data = [XMLPage(file, format_type).to_container() for file in evaluation_data]
        elif not format_type:
            pass
        else:
            raise ValueError(f'format_type {format_type} not in [alto, page, xml, None].')

        if not training_data:
            raise ValueError('No training data provided. Please add some.')

        transforms = ImageInputTransforms(batch,
                                          height,
                                          width,
                                          channels,
                                          self.hparams.hyper_params['padding'],
                                          valid_norm=False,
                                          force_binarization=force_binarization)

        self.example_input_array = torch.Tensor(batch,
                                                channels,
                                                height if height else 400,
                                                width if width else 300)

        # set multiprocessing tensor sharing strategy
        if 'file_system' in torch.multiprocessing.get_all_sharing_strategies():
            logger.debug('Setting multiprocessing tensor sharing strategy to file_system')
            torch.multiprocessing.set_sharing_strategy('file_system')

        if not valid_regions:
            valid_regions = None
        if not valid_baselines:
            valid_baselines = None

        if suppress_regions:
            valid_regions = []
            merge_regions = None
        if suppress_baselines:
            valid_baselines = []
            merge_baselines = None

        train_set = BaselineSet(line_width=self.hparams.hyper_params['line_width'],
                                im_transforms=transforms,
                                augmentation=self.hparams.hyper_params['augment'],
                                valid_baselines=valid_baselines,
                                merge_baselines=merge_baselines,
                                valid_regions=valid_regions,
                                merge_regions=merge_regions)

        for page in training_data:
            train_set.add(page)

        if evaluation_data:
            val_set = BaselineSet(line_width=self.hparams.hyper_params['line_width'],
                                  im_transforms=transforms,
                                  augmentation=False,
                                  valid_baselines=valid_baselines,
                                  merge_baselines=merge_baselines,
                                  valid_regions=valid_regions,
                                  merge_regions=merge_regions)

            for page in evaluation_data:
                val_set.add(page)

            train_set = Subset(train_set, range(len(train_set)))
            val_set = Subset(val_set, range(len(val_set)))
        else:
            train_len = int(len(train_set)*partition)
            val_len = len(train_set) - train_len
            logger.info(f'No explicit validation data provided. Splitting off '
                        f'{val_len} (of {len(train_set)}) samples to validation '
                        'set.')
            train_set, val_set = random_split(train_set, (train_len, val_len))

        if len(train_set) == 0:
            raise ValueError('No valid training data provided. Please add some.')

        if len(val_set) == 0:
            raise ValueError('No valid validation data provided. Please add some.')

        # overwrite class mapping in validation set
        val_set.dataset.num_classes = train_set.dataset.num_classes
        val_set.dataset.class_mapping = train_set.dataset.class_mapping

        self.train_set = train_set
        self.val_set = val_set

    def forward(self, x):
        return self.nn.nn(x)

    def training_step(self, batch, batch_idx):
        input, target = batch['image'], batch['target']
        output, _ = self.nn.nn(input)
        output = F.interpolate(output, size=(target.size(2), target.size(3)))
        loss = self.nn.criterion(output, target)
        self.log('train_loss', loss, on_step=True, on_epoch=False, prog_bar=False, logger=True)
        return loss

    def validation_step(self, batch, batch_idx):
        x, y = batch['image'], batch['target']
        pred, _ = self.nn.nn(x)
        # scale target to output size
        y = F.interpolate(y, size=(pred.size(2), pred.size(3))).int()

        self.val_px_accuracy.update(pred, y)
        self.val_mean_accuracy.update(pred, y)
        self.val_mean_iu.update(pred, y)
        self.val_freq_iu.update(pred, y)

    def on_validation_epoch_end(self):
        if not self.trainer.sanity_checking:
            pixel_accuracy = self.val_px_accuracy.compute()
            mean_accuracy = self.val_mean_accuracy.compute()
            mean_iu = self.val_mean_iu.compute()
            freq_iu = self.val_freq_iu.compute()

            if mean_iu > self.best_metric:
                logger.debug(f'Updating best metric from {self.best_metric} ({self.best_epoch}) to {mean_iu} ({self.current_epoch})')
                self.best_epoch = self.current_epoch
                self.best_metric = mean_iu

            logger.info(f'validation run: accuracy {pixel_accuracy} mean_acc {mean_accuracy} mean_iu {mean_iu} freq_iu {freq_iu}')

            self.log('val_accuracy', pixel_accuracy, on_step=False, on_epoch=True, prog_bar=True, logger=True)
            self.log('val_mean_acc', mean_accuracy, on_step=False, on_epoch=True, prog_bar=True, logger=True)
            self.log('val_mean_iu', mean_iu, on_step=False, on_epoch=True, prog_bar=True, logger=True)
            self.log('val_freq_iu', freq_iu, on_step=False, on_epoch=True, prog_bar=True, logger=True)
            self.log('val_metric', mean_iu, on_step=False, on_epoch=True, prog_bar=False, logger=True)

        # reset metrics even if sanity checking
        self.val_px_accuracy.reset()
        self.val_mean_accuracy.reset()
        self.val_mean_iu.reset()
        self.val_freq_iu.reset()

    def setup(self, stage: Optional[str] = None):
        # finalize models in case of appending/loading
        if stage in [None, 'fit']:
            if not self.model:
                self.spec = f'[{self.spec[1:-1]} O2l{self.train_set.dataset.num_classes}]'
                logger.info(f'Creating model {self.spec} with {self.train_set.dataset.num_classes} outputs')
                nn = vgsl.TorchVGSLModel(self.spec)
                if self.bounding_regions is not None:
                    nn.user_metadata['bounding_regions'] = self.bounding_regions
                nn.user_metadata['topline'] = self.topline
                self.nn = nn
            else:
                if self.train_set.dataset.class_mapping['baselines'].keys() != self.nn.user_metadata['class_mapping']['baselines'].keys() or \
                   self.train_set.dataset.class_mapping['regions'].keys() != self.nn.user_metadata['class_mapping']['regions'].keys():

                    bl_diff = set(self.train_set.dataset.class_mapping['baselines'].keys()).symmetric_difference(
                        set(self.nn.user_metadata['class_mapping']['baselines'].keys()))
                    regions_diff = set(self.train_set.dataset.class_mapping['regions'].keys()).symmetric_difference(
                        set(self.nn.user_metadata['class_mapping']['regions'].keys()))

                    if self.resize == 'fail':
                        raise ValueError(f'Training data and model class mapping differ (bl: {bl_diff}, regions: {regions_diff}')
                    elif self.resize == 'union':
                        new_bls = self.train_set.dataset.class_mapping['baselines'].keys() - self.nn.user_metadata['class_mapping']['baselines'].keys()
                        new_regions = self.train_set.dataset.class_mapping['regions'].keys() - self.nn.user_metadata['class_mapping']['regions'].keys()
                        cls_idx = max(max(self.nn.user_metadata['class_mapping']['baselines'].values()) if self.nn.user_metadata['class_mapping']['baselines'] else -1, # noqa
                                      max(self.nn.user_metadata['class_mapping']['regions'].values()) if self.nn.user_metadata['class_mapping']['regions'] else -1) # noqa
                        logger.info(f'Adding {len(new_bls) + len(new_regions)} missing types to network output layer.')
                        self.nn.resize_output(cls_idx + len(new_bls) + len(new_regions) + 1)
                        for c in new_bls:
                            cls_idx += 1
                            self.nn.user_metadata['class_mapping']['baselines'][c] = cls_idx
                        for c in new_regions:
                            cls_idx += 1
                            self.nn.user_metadata['class_mapping']['regions'][c] = cls_idx
                    elif self.resize == 'new':
                        logger.info('Fitting network exactly to training set.')
                        new_bls = self.train_set.dataset.class_mapping['baselines'].keys() - self.nn.user_metadata['class_mapping']['baselines'].keys()
                        new_regions = self.train_set.dataset.class_mapping['regions'].keys() - self.nn.user_metadata['class_mapping']['regions'].keys()
                        del_bls = self.nn.user_metadata['class_mapping']['baselines'].keys() - self.train_set.dataset.class_mapping['baselines'].keys()
                        del_regions = self.nn.user_metadata['class_mapping']['regions'].keys() - self.train_set.dataset.class_mapping['regions'].keys()

                        logger.info(f'Adding {len(new_bls) + len(new_regions)} missing '
                                    f'types and removing {len(del_bls) + len(del_regions)} to network output layer ')
                        cls_idx = max(max(self.nn.user_metadata['class_mapping']['baselines'].values()) if self.nn.user_metadata['class_mapping']['baselines'] else -1, # noqa
                                      max(self.nn.user_metadata['class_mapping']['regions'].values()) if self.nn.user_metadata['class_mapping']['regions'] else -1) # noqa

                        del_indices = [self.nn.user_metadata['class_mapping']['baselines'][x] for x in del_bls]
                        del_indices.extend(self.nn.user_metadata['class_mapping']['regions'][x] for x in del_regions)
                        self.nn.resize_output(cls_idx + len(new_bls) + len(new_regions) -
                                              len(del_bls) - len(del_regions) + 1, del_indices)

                        # delete old baseline/region types
                        cls_idx = min(min(self.nn.user_metadata['class_mapping']['baselines'].values()) if self.nn.user_metadata['class_mapping']['baselines'] else np.inf, # noqa
                                      min(self.nn.user_metadata['class_mapping']['regions'].values()) if self.nn.user_metadata['class_mapping']['regions'] else np.inf) # noqa

                        bls = {}
                        for k, v in sorted(self.nn.user_metadata['class_mapping']['baselines'].items(), key=lambda item: item[1]):
                            if k not in del_bls:
                                bls[k] = cls_idx
                                cls_idx += 1

                        regions = {}
                        for k, v in sorted(self.nn.user_metadata['class_mapping']['regions'].items(), key=lambda item: item[1]):
                            if k not in del_regions:
                                regions[k] = cls_idx
                                cls_idx += 1

                        self.nn.user_metadata['class_mapping']['baselines'] = bls
                        self.nn.user_metadata['class_mapping']['regions'] = regions

                        # add new baseline/region types
                        cls_idx -= 1
                        for c in new_bls:
                            cls_idx += 1
                            self.nn.user_metadata['class_mapping']['baselines'][c] = cls_idx
                        for c in new_regions:
                            cls_idx += 1
                            self.nn.user_metadata['class_mapping']['regions'][c] = cls_idx
                    else:
                        raise ValueError(f'invalid resize parameter value {self.resize}')
                # backfill train_set/val_set mapping if key-equal as the actual
                # numbering in the train_set might be different
                self.train_set.dataset.class_mapping = self.nn.user_metadata['class_mapping']
                self.val_set.dataset.class_mapping = self.nn.user_metadata['class_mapping']

            # updates model's hyper params with user-defined ones
            self.nn.hyper_params = self.hparams.hyper_params

            # change topline/baseline switch
            loc = {None: 'centerline',
                   True: 'topline',
                   False: 'baseline'}

            if 'topline' not in self.nn.user_metadata:
                logger.warning(f'Setting baseline location to {loc[self.topline]} from unset model.')
            elif self.nn.user_metadata['topline'] != self.topline:
                from_loc = loc[self.nn.user_metadata['topline']]
                logger.warning(f'Changing baseline location from {from_loc} to {loc[self.topline]}.')
            self.nn.user_metadata['topline'] = self.topline

            logger.info('Training line types:')
            for k, v in self.train_set.dataset.class_mapping['baselines'].items():
                logger.info(f'  {k}\t{v}\t{self.train_set.dataset.class_stats["baselines"][k]}')
            logger.info('Training region types:')
            for k, v in self.train_set.dataset.class_mapping['regions'].items():
                logger.info(f'  {k}\t{v}\t{self.train_set.dataset.class_stats["regions"][k]}')

            if len(self.train_set) == 0:
                raise ValueError('No valid training data was provided to the train command. Please add valid XML data.')

            # set model type metadata field and dump class_mapping
            self.nn.model_type = 'segmentation'
            self.nn.user_metadata['class_mapping'] = self.val_set.dataset.class_mapping

            # for model size/trainable parameter output
            self.net = self.nn.nn

            torch.set_num_threads(max(self.num_workers, 1))

            # set up validation metrics after output classes have been determined
            self.val_px_accuracy = MultilabelAccuracy(average='micro', num_labels=self.train_set.dataset.num_classes)
            self.val_mean_accuracy = MultilabelAccuracy(average='macro', num_labels=self.train_set.dataset.num_classes)
            self.val_mean_iu = MultilabelJaccardIndex(average='macro', num_labels=self.train_set.dataset.num_classes)
            self.val_freq_iu = MultilabelJaccardIndex(average='weighted', num_labels=self.train_set.dataset.num_classes)

    def train_dataloader(self):
        return DataLoader(self.train_set,
                          batch_size=1,
                          num_workers=self.num_workers,
                          shuffle=True,
                          pin_memory=True)

    def val_dataloader(self):
        return DataLoader(self.val_set,
                          shuffle=False,
                          batch_size=1,
                          num_workers=self.num_workers,
                          pin_memory=True)

    def configure_callbacks(self):
        callbacks = []
        if self.hparams.hyper_params['quit'] == 'early':
            callbacks.append(EarlyStopping(monitor='val_mean_iu',
                                           mode='max',
                                           patience=self.hparams.hyper_params['lag'],
                                           stopping_threshold=1.0))

        return callbacks

    # configuration of optimizers and learning rate schedulers
    # --------------------------------------------------------
    #
    # All schedulers are created internally with a frequency of step to enable
    # batch-wise learning rate warmup. In lr_scheduler_step() calls to the
    # scheduler are then only performed at the end of the epoch.
    def configure_optimizers(self):
        return _configure_optimizer_and_lr_scheduler(self.hparams.hyper_params,
                                                     self.nn.nn.parameters(),
                                                     len_train_set=len(self.train_set),
                                                     loss_tracking_mode='max')

    def optimizer_step(self, epoch, batch_idx, optimizer, optimizer_closure):
        # update params
        optimizer.step(closure=optimizer_closure)

        # linear warmup between 0 and the initial learning rate `lrate` in `warmup`
        # steps.
        if self.hparams.hyper_params['warmup'] and self.trainer.global_step < self.hparams.hyper_params['warmup']:
            lr_scale = min(1.0, float(self.trainer.global_step + 1) / self.hparams.hyper_params['warmup'])
            for pg in optimizer.param_groups:
                pg["lr"] = lr_scale * self.hparams.hyper_params['lrate']

    def lr_scheduler_step(self, scheduler, metric):
        if not self.hparams.hyper_params['warmup'] or self.trainer.global_step >= self.hparams.hyper_params['warmup']:
            # step OneCycleLR each batch if not in warmup phase
            if isinstance(scheduler, lr_scheduler.OneCycleLR):
                scheduler.step()
            # step every other scheduler epoch-wise
            elif self.trainer.is_last_batch:
                scheduler.step()


def _configure_optimizer_and_lr_scheduler(hparams, params, len_train_set=None, loss_tracking_mode='max'):
    optimizer = hparams.get("optimizer")
    lrate = hparams.get("lrate")
    momentum = hparams.get("momentum")
    weight_decay = hparams.get("weight_decay")
    schedule = hparams.get("schedule")
    gamma = hparams.get("gamma")
    cos_t_max = hparams.get("cos_t_max")
    cos_min_lr = hparams.get("cos_min_lr")
    step_size = hparams.get("step_size")
    rop_factor = hparams.get("rop_factor")
    rop_patience = hparams.get("rop_patience")
    epochs = hparams.get("epochs")
    completed_epochs = hparams.get("completed_epochs")

    # XXX: Warmup is not configured here because it needs to be manually done in optimizer_step()
    logger.debug(f'Constructing {optimizer} optimizer (lr: {lrate}, momentum: {momentum})')
    if optimizer == 'Adam':
        optim = torch.optim.Adam(params, lr=lrate, weight_decay=weight_decay)
    else:
        optim = getattr(torch.optim, optimizer)(params,
                                                lr=lrate,
                                                momentum=momentum,
                                                weight_decay=weight_decay)
    lr_sched = {}
    if schedule == 'exponential':
        lr_sched = {'scheduler': lr_scheduler.ExponentialLR(optim, gamma, last_epoch=completed_epochs-1),
                    'interval': 'step'}
    elif schedule == 'cosine':
        lr_sched = {'scheduler': lr_scheduler.CosineAnnealingLR(optim,
                                                                cos_t_max,
                                                                cos_min_lr,
                                                                last_epoch=completed_epochs-1),
                    'interval': 'step'}
    elif schedule == 'step':
        lr_sched = {'scheduler': lr_scheduler.StepLR(optim, step_size, gamma, last_epoch=completed_epochs-1),
                    'interval': 'step'}
    elif schedule == 'reduceonplateau':
        lr_sched = {'scheduler': lr_scheduler.ReduceLROnPlateau(optim,
                                                                mode=loss_tracking_mode,
                                                                factor=rop_factor,
                                                                patience=rop_patience),
                    'interval': 'step'}
    elif schedule == '1cycle':
        if epochs <= 0:
            raise ValueError('1cycle learning rate scheduler selected but '
                             'number of epochs is less than 0 '
                             f'({epochs}).')
        last_epoch = completed_epochs*len_train_set if completed_epochs else -1
        lr_sched = {'scheduler': lr_scheduler.OneCycleLR(optim,
                                                         max_lr=lrate,
                                                         epochs=epochs,
                                                         steps_per_epoch=len_train_set,
                                                         last_epoch=last_epoch),
                    'interval': 'step'}
    elif schedule != 'constant':
        raise ValueError(f'Unsupported learning rate scheduler {schedule}.')

    ret = {'optimizer': optim}
    if lr_sched:
        ret['lr_scheduler'] = lr_sched

    if schedule == 'reduceonplateau':
        lr_sched['monitor'] = 'val_metric'
        lr_sched['strict'] = False
        lr_sched['reduce_on_plateau'] = True

    return ret



# -*- coding: utf-8 -*-
import unittest
from pathlib import Path

from PIL import Image
from pytest import raises

from kraken.lib.exceptions import KrakenInputException
from kraken.pageseg import segment

thisfile = Path(__file__).resolve().parent
resources = thisfile / 'resources'

class TestPageSeg(unittest.TestCase):

    """
    Tests of the page segmentation functionality
    """
    def test_segment_color(self):
        """
        Test correct handling of color input.
        """
        with raises(KrakenInputException):
            with Image.open(resources / 'input.jpg') as im:
                segment(im)

    def test_segment_bw(self):
        """
        Tests segmentation of bi-level input.
        """
        with Image.open(resources / 'bw.png') as im:
            seg = segment(im)
            self.assertEqual(seg.type, 'bbox')
            # test if line count is roughly correct
            self.assertAlmostEqual(len(seg.lines), 30, msg='Segmentation differs '
                                   'wildly from true line count', delta=5)
            # check if lines do not extend beyond image
            for line in seg.lines:
                box = line.bbox
                self.assertLess(0, box[0], msg='Line x0 < 0')
                self.assertLess(0, box[1], msg='Line y0 < 0')
                self.assertGreater(im.size[0], box[2], msg='Line x1 > {}'.format(im.size[0]))
                self.assertGreater(im.size[1], box[3], msg='Line y1 > {}'.format(im.size[1]))



#
# Copyright 2020 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
Default VGSL specs and hyperparameters
"""

SEGMENTATION_SPEC = '[1,1800,0,3 Cr7,7,64,2,2 Gn32 Cr3,3,128,2,2 Gn32 Cr3,3,128 Gn32 Cr3,3,256 Gn32 Cr3,3,256 Gn32 Lbx32 Lby32 Cr1,1,32 Gn32 Lby32 Lbx32]' # NOQA
RECOGNITION_SPEC = '[1,120,0,1 Cr3,13,32 Do0.1,2 Mp2,2 Cr3,13,32 Do0.1,2 Mp2,2 Cr3,9,64 Do0.1,2 Mp2,2 Cr3,9,64 Do0.1,2 S1(1x0)1,3 Lbx200 Do0.1,2 Lbx200 Do0.1,2 Lbx200 Do]' # NOQA

READING_ORDER_HYPER_PARAMS = {'lrate': 0.001,
                              'freq': 1.0,
                              'batch_size': 15000,
                              'epochs': 3000,
                              'lag': 300,
                              'min_delta': None,
                              'quit': 'early',
                              'optimizer': 'Adam',
                              'momentum': 0.9,
                              'weight_decay': 0.01,
                              'schedule': 'cosine',
                              'completed_epochs': 0,
                              # lr scheduler params
                              # step/exp decay
                              'step_size': 10,
                              'gamma': 0.1,
                              # reduce on plateau
                              'rop_factor': 0.1,
                              'rop_patience': 5,
                              # cosine
                              'cos_t_max': 100,
                              'cos_min_lr': 0.001,
                              'warmup': 0,
                              }

RECOGNITION_PRETRAIN_HYPER_PARAMS = {'pad': 16,
                                     'freq': 1.0,
                                     'batch_size': 64,
                                     'quit': 'early',
                                     'epochs': -1,
                                     'min_epochs': 100,
                                     'lag': 5,
                                     'min_delta': None,
                                     'optimizer': 'Adam',
                                     'lrate': 1e-6,
                                     'momentum': 0.9,
                                     'weight_decay': 0.01,
                                     'schedule': 'cosine',
                                     'completed_epochs': 0,
                                     'augment': False,
                                     # lr scheduler params
                                     # step/exp decay
                                     'step_size': 10,
                                     'gamma': 0.1,
                                     # reduce on plateau
                                     'rop_factor': 0.1,
                                     'rop_patience': 5,
                                     # cosine
                                     'cos_t_max': 100,
                                     'cos_min_lr': 1e-7,
                                     # masking parameters
                                     'mask_width': 4,
                                     'mask_prob': 0.5,
                                     'num_negatives': 100,
                                     'logit_temp': 0.1,
                                     'warmup': 32000,
                                     }

RECOGNITION_HYPER_PARAMS = {'pad': 16,
                            'freq': 1.0,
                            'batch_size': 1,
                            'quit': 'early',
                            'epochs': -1,
                            'min_epochs': 0,
                            'lag': 10,
                            'min_delta': None,
                            'optimizer': 'Adam',
                            'lrate': 1e-3,
                            'momentum': 0.9,
                            'weight_decay': 0.0,
                            'schedule': 'constant',
                            'normalization': None,
                            'normalize_whitespace': True,
                            'completed_epochs': 0,
                            'augment': False,
                            # lr scheduler params
                            # step/exp decay
                            'step_size': 10,
                            'gamma': 0.1,
                            # reduce on plateau
                            'rop_factor': 0.1,
                            'rop_patience': 5,
                            # cosine
                            'cos_t_max': 50,
                            'cos_min_lr': 1e-4,
                            'warmup': 0,
                            'freeze_backbone': 0,
                            }

SEGMENTATION_HYPER_PARAMS = {'line_width': 8,
                             'padding': (0, 0),
                             'freq': 1.0,
                             'quit': 'fixed',
                             'epochs': 50,
                             'min_epochs': 0,
                             'lag': 10,
                             'min_delta': None,
                             'optimizer': 'Adam',
                             'lrate': 2e-4,
                             'momentum': 0.9,
                             'weight_decay': 1e-5,
                             'schedule': 'constant',
                             'completed_epochs': 0,
                             'augment': False,
                             # lr scheduler params
                             # step/exp decay
                             'step_size': 10,
                             'gamma': 0.1,
                             # reduce on plateau
                             'rop_factor': 0.1,
                             'rop_patience': 5,
                             # cosine
                             'cos_t_max': 50,
                             'cos_min_lr': 2e-5,
                             'warmup': 0,
                             }



# -*- coding: utf-8 -*-
import unittest
from typing import Sequence, Tuple

import numpy as np
import pytest
import shapely.geometry as geom

from kraken.lib.segmentation import is_in_region, reading_order, topsort


def polygon_slices(polygon: Sequence[Tuple[int, int]]) -> Tuple[slice, slice]:
    """Convert polygons to slices for reading_order"""
    linestr = geom.LineString(polygon)
    slices = (slice(linestr.bounds[1], linestr.bounds[3]),
             slice(linestr.bounds[0], linestr.bounds[2]))
    return slices


class TestReadingOrder(unittest.TestCase):

    """
    Test the reading order algorithms.
    """
    def test_is_in_region(self):
        """
        A line should be in its rectangular bounding box.
        """
        line = geom.LineString([(0, 0), (1, 1)])
        polygon = geom.Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])
        self.assertTrue(is_in_region(line, polygon))

    def test_is_in_region2(self):
        """
        A real baseline should be in its polygonization.
        """
        line = geom.LineString([(268, 656), (888, 656)])
        polygon = geom.Polygon([(268, 656), (265, 613), (885, 611), (888, 656), (885, 675), (265, 672)])
        self.assertTrue(is_in_region(line, polygon))

    def test_is_in_region3(self):
        """
        A line that does not cross the box should not be in the region.
        """
        line = geom.LineString([(2, 2), (1, 1)])
        polygon = geom.Polygon([(0, 0), (1, 0), (1, 1), (0, 1)])
        self.assertFalse(is_in_region(line, polygon))

    def test_order_simple_over_under(self):
        """
        Two lines (as their polygonal boundaries) are already in order.
        In this example, the boundaries are rectangles that align vertically,
        have horizontal base lines and do not overlap or touch::

            AAAA

            BBBB

        The reading order should be the same for left-to-right and right-to-left.
        """
        polygon0 = [[10, 10], [10, 20], [100, 20], [100, 10], [10, 10]]
        polygon1 = [[10, 30], [10, 40], [100, 40], [100, 30], [10, 30]]
        order_lr = reading_order([polygon_slices(line) for line in [polygon0, polygon1]])
        order_rl = reading_order([polygon_slices(line) for line in [polygon0, polygon1]], 'rl')
        # line0 should come before line1, lines do not come before themselves
        expected = np.array([[0, 1], [0, 0]])
        self.assertTrue(np.array_equal(order_lr, expected), "Reading order is not as expected: {}".format(order_lr))
        self.assertTrue(np.array_equal(order_rl, expected), "Reading order is not as expected: {}".format(order_rl))

    def test_order_simple_over_under_touching(self):
        """
        Two lines (as their polygonal boundaries) are already in order.
        In this example, the boundaries are rectangles that align vertically,
        have horizontal base lines and touch::

            AAAA
            BBBB

        The reading order should be the same for left-to-right and right-to-left.
        """
        polygon0 = [[10, 10], [10, 30], [100, 30], [100, 10], [10, 10]]
        polygon1 = [[10, 30], [10, 40], [100, 40], [100, 30], [10, 30]]
        order_lr = reading_order([polygon_slices(line) for line in [polygon0, polygon1]])
        order_rl = reading_order([polygon_slices(line) for line in [polygon0, polygon1]], 'rl')
        # line0 should come before line1, lines do not come before themselves
        expected = np.array([[0, 1], [0, 0]])
        self.assertTrue(np.array_equal(order_lr, expected), "Reading order is not as expected: {}".format(order_lr))
        self.assertTrue(np.array_equal(order_rl, expected), "Reading order is not as expected: {}".format(order_rl))

    def test_order_simple_left_right(self):
        """
        Two lines (as their polygonal boundaries) are already in order.
        In this example, the boundaries are rectangles that align horizontally,
        have horizontal base lines and do not overlap or touch::

            AAAA  BBBB

        """
        polygon0 = [[10, 10], [10, 20], [100, 20], [100, 10], [10, 10]]
        polygon1 = [[150, 10], [150, 20], [250, 20], [250, 10], [150, 10]]
        order = reading_order([polygon_slices(line) for line in [polygon0, polygon1]])
        # line0 should come before line1, lines do not come before themselves
        expected = np.array([[0, 1], [0, 0]])
        self.assertTrue(np.array_equal(order, expected), "Reading order is not as expected: {}".format(order))

    @pytest.mark.xfail
    def test_order_simple_left_right_touching(self):
        """
        Two lines (as their polygonal boundaries) are already in order.
        In this example, the boundaries are rectangles that align horizontally,
        have horizontal base lines and touch::

            AAAABBBB

        """
        polygon0 = [[10, 10], [10, 20], [100, 20], [100, 10], [10, 10]]
        polygon1 = [[100, 10], [100, 20], [250, 20], [250, 10], [100, 10]]
        order = reading_order([polygon_slices(line) for line in [polygon0, polygon1]])
        # line0 should come before line1, lines do not come before themselves
        expected = np.array([[0, 1], [0, 0]])
        self.assertTrue(np.array_equal(order, expected), "Reading order is not as expected: {}".format(order))

    def test_order_simple_right_left(self):
        """
        Two lines (as their polygonal boundaries) are in reverse RTL-order.
        In this example, the boundaries are rectangles that align horizontally,
        have horizontal base lines and do not overlap or touch::

            BBBB  AAAA

        """
        polygon0 = [[10, 10], [10, 20], [100, 20], [100, 10], [10, 10]]
        polygon1 = [[150, 10], [150, 20], [250, 20], [250, 10], [150, 10]]
        order = reading_order([polygon_slices(line) for line in [polygon0, polygon1]], 'rl')
        # line1 should come before line0, lines do not come before themselves
        expected = np.array([[0, 0], [1, 0]])
        self.assertTrue(np.array_equal(order, expected), "Reading order is not as expected: {}".format(order))

    @pytest.mark.xfail
    def test_order_simple_right_left_touching(self):
        """
        Two lines (as their polygonal boundaries) are in reverse RTL-order.
        In this example, the boundaries are rectangles that align horizontally,
        have horizontal base lines and touch::

            BBBBAAAA

        """
        polygon0 = [[10, 10], [10, 20], [100, 20], [100, 10], [10, 10]]
        polygon1 = [[100, 10], [100, 20], [250, 20], [250, 10], [100, 10]]
        order = reading_order([polygon_slices(line) for line in [polygon0, polygon1]])
        # line1 should come before line0, lines do not come before themselves
        expected = np.array([[0, 0], [1, 0]])
        self.assertTrue(np.array_equal(order, expected), "Reading order is not as expected: {}".format(order))

    def test_order_real_reverse(self):
        """
        Real example: lines are in reverse order.
        The reading order should be the same for left-to-right and right-to-left.
        """
        polygon0 = [[474, 2712], [466, 2669], [1741, 2655], [1749, 2696], [1746, 2709], [474, 2725]]
        polygon1 = [[493, 2409], [488, 2374], [1733, 2361], [1741, 2395], [1738, 2409], [493, 2422]]
        order_lr = reading_order([polygon_slices(line) for line in [polygon0, polygon1]])
        order_rl = reading_order([polygon_slices(line) for line in [polygon0, polygon1]], 'rl')
        # line1 should come before line0, lines do not come before themselves
        expected = np.array([[0, 0], [1, 0]])
        self.assertTrue(np.array_equal(order_lr, expected), "Reading order is not as expected: {}".format(order_lr))
        self.assertTrue(np.array_equal(order_rl, expected), "Reading order is not as expected: {}".format(order_rl))

    def test_order_real_in_order(self):
        """
        Real (modified) example: lines are in order.
        The reading order should be the same for left-to-right and right-to-left.
        """
        polygon0 = [[493, 2409], [488, 2374], [1733, 2361], [1741, 2395], [1738, 2409], [493, 2422]]
        polygon1 = [[474, 2712], [466, 2669], [1741, 2655], [1749, 2696], [1746, 2709], [474, 2725]]
        order_lr = reading_order([polygon_slices(line) for line in [polygon0, polygon1]])
        order_rl = reading_order([polygon_slices(line) for line in [polygon0, polygon1]], 'rl')
        # line0 should come before line1, lines do not come before themselves
        expected = np.array([[0, 1], [0, 0]])
        self.assertTrue(np.array_equal(order_lr, expected), "Reading order is not as expected: {}".format(order_lr))
        self.assertTrue(np.array_equal(order_rl, expected), "Reading order is not as expected: {}".format(order_rl))

    def test_topsort_ordered(self):
        """
        Return list for three lines that are already in order.
        """
        partial_sort = np.array([[1, 1, 1], [0, 1, 1], [0, 0, 1]])
        expected = [0, 1, 2]
        self.assertTrue(np.array_equal(topsort(partial_sort), expected))

    def test_topsort_ordered_no_self(self):
        """
        Return list for three lines that are already in order.
        """
        partial_sort = np.array([[0, 1, 1], [0, 0, 1], [0, 0, 0]])
        expected = [0, 1, 2]
        self.assertTrue(np.array_equal(topsort(partial_sort), expected))

    def test_topsort_unordered(self):
        """
        Return list for three lines that are partially in order.
        """
        partial_sort = np.array([[1, 1, 1], [0, 1, 0], [0, 1, 1]])
        expected = [0, 2, 1]
        self.assertTrue(np.array_equal(topsort(partial_sort), expected))

    def test_topsort_unordered_no_self(self):
        """
        Return list for three lines that are partially in order.
        """
        partial_sort = np.array([[0, 1, 1], [0, 0, 0], [0, 1, 0]])
        expected = [0, 2, 1]
        self.assertTrue(np.array_equal(topsort(partial_sort), expected))



#
# Copyright 2022 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
kraken.ketos.repo
~~~~~~~~~~~~~~~~~

Command line driver for publishing models to the model repository.
"""
import logging
import os

import click

from .util import message

logging.captureWarnings(True)
logger = logging.getLogger('kraken')


@click.command('publish')
@click.pass_context
@click.option('-i', '--metadata', show_default=True,
              type=click.File(mode='r', lazy=True), help='Metadata for the '
              'model. Will be prompted from the user if not given')
@click.option('-a', '--access-token', prompt=True, help='Zenodo access token')
@click.option('-p', '--private/--public', default=False, help='Disables Zenodo '
              'community inclusion request. Allows upload of models that will not show '
              'up on `kraken list` output')
@click.argument('model', nargs=1, type=click.Path(exists=False, readable=True, dir_okay=False))
def publish(ctx, metadata, access_token, private, model):
    """
    Publishes a model on the zenodo model repository.
    """
    import json

    import importlib_resources
    from jsonschema import validate
    from jsonschema.exceptions import ValidationError

    from kraken import repo
    from kraken.lib import models
    from kraken.lib.progress import KrakenDownloadProgressBar

    ref = importlib_resources.files('kraken').joinpath('metadata.schema.json')
    with open(ref, 'rb') as fp:
        schema = json.load(fp)

    nn = models.load_any(model)

    if not metadata:
        author = click.prompt('author')
        affiliation = click.prompt('affiliation')
        summary = click.prompt('summary')
        description = click.edit('Write long form description (training data, transcription standards) of the model here')
        accuracy_default = None
        # take last accuracy measurement in model metadata
        if 'accuracy' in nn.nn.user_metadata and nn.nn.user_metadata['accuracy']:
            accuracy_default = nn.nn.user_metadata['accuracy'][-1][1] * 100
        accuracy = click.prompt('accuracy on test set', type=float, default=accuracy_default)
        script = [
            click.prompt(
                'script',
                type=click.Choice(
                    sorted(
                        schema['properties']['script']['items']['enum'])),
                show_choices=True)]
        license = click.prompt(
            'license',
            type=click.Choice(
                sorted(
                    schema['properties']['license']['enum'])),
            show_choices=True)
        metadata = {
            'authors': [{'name': author, 'affiliation': affiliation}],
            'summary': summary,
            'description': description,
            'accuracy': accuracy,
            'license': license,
            'script': script,
            'name': os.path.basename(model),
            'graphemes': ['a']
        }
        while True:
            try:
                validate(metadata, schema)
            except ValidationError as e:
                message(e.message)
                metadata[e.path[-1]] = click.prompt(e.path[-1], type=float if e.schema['type'] == 'number' else str)
                continue
            break

    else:
        metadata = json.load(metadata)
        validate(metadata, schema)
    metadata['graphemes'] = [char for char in ''.join(nn.codec.c2l.keys())]
    with KrakenDownloadProgressBar() as progress:
        upload_task = progress.add_task('Uploading', total=0, visible=True if not ctx.meta['verbose'] else False)
        oid = repo.publish_model(model, metadata, access_token, lambda total, advance: progress.update(upload_task, total=total, advance=advance), private)
    message('model PID: {}'.format(oid))



#
# Copyright 2015 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
Utility functions for data loading and training of VGSL networks.
"""
import traceback
from collections import defaultdict
from itertools import groupby
from typing import TYPE_CHECKING, Any, Callable, Dict, Sequence, Tuple

import numpy as np
import shapely.geometry as geom
import torch
import torch.nn.functional as F
from PIL import Image
from shapely.ops import split
from skimage.draw import polygon
from torch.utils.data import Dataset
from torchvision import transforms

from kraken.lib.segmentation import scale_regions

if TYPE_CHECKING:
    from kraken.containers import Segmentation


__all__ = ['BaselineSet']

import logging

logger = logging.getLogger(__name__)


class BaselineSet(Dataset):
    """
    Dataset for training a baseline/region segmentation model.
    """
    def __init__(self,
                 line_width: int = 4,
                 padding: Tuple[int, int, int, int] = (0, 0, 0, 0),
                 im_transforms: Callable[[Any], torch.Tensor] = transforms.Compose([]),
                 augmentation: bool = False,
                 valid_baselines: Sequence[str] = None,
                 merge_baselines: Dict[str, Sequence[str]] = None,
                 valid_regions: Sequence[str] = None,
                 merge_regions: Dict[str, Sequence[str]] = None):
        """
        Creates a dataset for a text-line and region segmentation model.

        Args:
            line_width: Height of the baseline in the scaled input.
            padding: Tuple of ints containing the left/right, top/bottom
                     padding of the input images.
            target_size: Target size of the image as a (height, width) tuple.
            augmentation: Enable/disable augmentation.
            valid_baselines: Sequence of valid baseline identifiers. If `None`
                             all are valid.
            merge_baselines: Sequence of baseline identifiers to merge.  Note
                             that merging occurs after entities not in valid_*
                             have been discarded.
            valid_regions: Sequence of valid region identifiers. If `None` all
                           are valid.
            merge_regions: Sequence of region identifiers to merge. Note that
                           merging occurs after entities not in valid_* have
                           been discarded.
        """
        super().__init__()
        self.imgs = []
        self.im_mode = '1'
        self.pad = padding
        self.targets = []
        # n-th entry contains semantic of n-th class
        self.class_mapping = {'aux': {'_start_separator': 0, '_end_separator': 1}, 'baselines': {}, 'regions': {}}
        # keep track of samples that failed to load
        self.failed_samples = set()
        self.class_stats = {'baselines': defaultdict(int), 'regions': defaultdict(int)}
        self.num_classes = 2
        self.mbl_dict = merge_baselines if merge_baselines is not None else {}
        self.mreg_dict = merge_regions if merge_regions is not None else {}
        self.valid_baselines = valid_baselines
        self.valid_regions = valid_regions

        self.aug = None
        if augmentation:
            import cv2
            cv2.setNumThreads(0)
            from albumentations import (Blur, Compose, ElasticTransform,
                                        HueSaturationValue, MedianBlur,
                                        MotionBlur, OneOf, OpticalDistortion,
                                        ShiftScaleRotate, ToFloat)

            self.aug = Compose([
                                ToFloat(),
                                OneOf([
                                    MotionBlur(p=0.2),
                                    MedianBlur(blur_limit=3, p=0.1),
                                    Blur(blur_limit=3, p=0.1),
                                ], p=0.2),
                                ShiftScaleRotate(shift_limit=0.0625, scale_limit=0.2, rotate_limit=45, p=0.2),
                                OneOf([
                                    OpticalDistortion(p=0.3),
                                    ElasticTransform(p=0.1),
                                ], p=0.2),
                                HueSaturationValue(hue_shift_limit=20, sat_shift_limit=0.1, val_shift_limit=0.1, p=0.3),
                               ], p=0.5)
        self.line_width = line_width
        self.transforms = im_transforms
        self.seg_type = None

    def add(self, doc: 'Segmentation'):
        """
        Adds a page to the dataset.

        Args:
            doc: A Segmentation container class.
        """
        if doc.type != 'baselines':
            raise ValueError(f'{doc} is of type {doc.type}. Expected "baselines".')

        baselines_ = defaultdict(list)
        for line in doc.lines:
            if self.valid_baselines is None or set(line.tags.values()).intersection(self.valid_baselines):
                tags = set(line.tags.values()).intersection(self.valid_baselines) if self.valid_baselines else line.tags.values()
                tags = set([self.mbl_dict.get(v, v) for v in tags])
                for tag in tags:
                    baselines_[tag].append(line.baseline)
                    self.class_stats['baselines'][tag] += 1

                    if tag not in self.class_mapping['baselines']:
                        self.num_classes += 1
                        self.class_mapping['baselines'][tag] = self.num_classes - 1

        regions_ = defaultdict(list)
        for k, v in doc.regions.items():
            if self.valid_regions is None or k in self.valid_regions:
                reg_type = self.mreg_dict.get(k, k)
                regions_[reg_type].extend(v)
                self.class_stats['baselines'][reg_type] += len(v)
                if reg_type not in self.class_mapping['regions']:
                    self.num_classes += 1
                    self.class_mapping['regions'][reg_type] = self.num_classes - 1
                self.class_stats['regions'][reg_type] += 1
        self.targets.append({'baselines': baselines_, 'regions': regions_})
        self.imgs.append(doc.imagename)

    def __getitem__(self, idx):
        im = self.imgs[idx]
        target = self.targets[idx]
        if not isinstance(im, Image.Image):
            try:
                logger.debug(f'Attempting to load {im}')
                im = Image.open(im)
                im, target = self.transform(im, target)
                return {'image': im, 'target': target}
            except Exception:
                self.failed_samples.add(idx)
                idx = np.random.randint(0, len(self.imgs))
                logger.debug(traceback.format_exc())
                logger.info(f'Failed. Replacing with sample {idx}')
                return self[idx]
        im, target = self.transform(im, target)
        return {'image': im, 'target': target}

    @staticmethod
    def _get_ortho_line(lineseg, point, line_width, offset):
        lineseg = np.array(lineseg)
        norm_vec = lineseg[1, ...] - lineseg[0, ...]
        norm_vec_len = np.sqrt(np.sum(norm_vec**2))
        unit_vec = norm_vec / norm_vec_len
        ortho_vec = unit_vec[::-1] * ((1, -1), (-1, 1))
        if offset == 'l':
            point -= unit_vec * line_width
        else:
            point += unit_vec * line_width
        return (ortho_vec * 10 + point).astype('int').tolist()

    def transform(self, image, target):
        orig_size = image.size
        image = self.transforms(image)
        scale = (image.shape[2] - 2*self.pad[1])/orig_size[0]
        t = torch.zeros((self.num_classes,) + tuple(np.subtract(image.shape[1:], (2*self.pad[1], 2*self.pad[0]))))
        start_sep_cls = self.class_mapping['aux']['_start_separator']
        end_sep_cls = self.class_mapping['aux']['_end_separator']

        for key, lines in target['baselines'].items():
            try:
                cls_idx = self.class_mapping['baselines'][key]
            except KeyError:
                # skip lines of classes not present in the training set
                continue
            for line in lines:
                # buffer out line to desired width
                line = [k for k, g in groupby(line)]
                line = np.array(line)*scale
                shp_line = geom.LineString(line)
                split_offset = min(5, shp_line.length/2)
                line_pol = np.array(shp_line.buffer(self.line_width/2, cap_style=2).boundary.coords, dtype=int)
                rr, cc = polygon(line_pol[:, 1], line_pol[:, 0], shape=image.shape[1:])
                t[cls_idx, rr, cc] = 1
                split_pt = shp_line.interpolate(split_offset).buffer(0.001)
                # top
                start_sep = np.array((split(shp_line, split_pt).geoms[0].buffer(self.line_width,
                                                                                cap_style=3).boundary.coords), dtype=int)
                rr_s, cc_s = polygon(start_sep[:, 1], start_sep[:, 0], shape=image.shape[1:])
                t[start_sep_cls, rr_s, cc_s] = 1
                t[start_sep_cls, rr, cc] = 0
                split_pt = shp_line.interpolate(-split_offset).buffer(0.001)
                # top
                end_sep = np.array((split(shp_line, split_pt).geoms[-1].buffer(self.line_width,
                                                                               cap_style=3).boundary.coords), dtype=int)
                rr_s, cc_s = polygon(end_sep[:, 1], end_sep[:, 0], shape=image.shape[1:])
                t[end_sep_cls, rr_s, cc_s] = 1
                t[end_sep_cls, rr, cc] = 0
        for key, regions in target['regions'].items():
            try:
                cls_idx = self.class_mapping['regions'][key]
            except KeyError:
                # skip regions of classes not present in the training set
                continue
            for region in regions:
                region = np.array(scale_regions([region.boundary], scale)[0])
                rr, cc = polygon(region[:, 1], region[:, 0], shape=image.shape[1:])
                t[cls_idx, rr, cc] = 1
        target = F.pad(t, self.pad)
        if self.aug:
            image = image.permute(1, 2, 0).numpy()
            target = target.permute(1, 2, 0).numpy()
            o = self.aug(image=image, mask=target)
            image = torch.tensor(o['image']).permute(2, 0, 1)
            target = torch.tensor(o['mask']).permute(2, 0, 1)
        return image, target

    def __len__(self):
        return len(self.imgs)



# -*- coding: utf-8 -*-
import os
import tempfile
import unittest

import torch
from pytest import raises

from kraken.lib import layers, vgsl


class TestVGSL(unittest.TestCase):
    """
    Testing VGSL module
    """
    def test_helper_train(self):
        """
        Tests train/eval mode helper methods
        """
        rnn = vgsl.TorchVGSLModel('[1,1,0,48 Lbx10 Do O1c57]')
        rnn.train()
        self.assertTrue(torch.is_grad_enabled())
        self.assertTrue(rnn.nn.training)
        rnn.eval()
        self.assertFalse(torch.is_grad_enabled())
        self.assertFalse(rnn.nn.training)

    @unittest.skip('works randomly on ci')
    def test_helper_threads(self):
        """
        Test openmp threads helper method.
        """
        rnn = vgsl.TorchVGSLModel('[1,1,0,48 Lbx10 Do O1c57]')
        rnn.set_num_threads(4)
        self.assertEqual(torch.get_num_threads(), 4)

    def test_save_model(self):
        """
        Test model serialization.
        """
        rnn = vgsl.TorchVGSLModel('[1,1,0,48 Lbx10 Do O1c57]')
        with tempfile.TemporaryDirectory() as dir:
            rnn.save_model(dir + '/foo.mlmodel')
            self.assertTrue(os.path.exists(dir + '/foo.mlmodel'))

    def test_append(self):
        """
        Test appending one VGSL spec to another.
        """
        rnn = vgsl.TorchVGSLModel('[1,1,0,48 Lbx10 Do O1c57]')
        rnn.append(1, '[Cr1,1,2 Gn2 Cr3,3,4]')
        self.assertEqual(rnn.spec, '[1,1,0,48 Lbx{L_0}10 Cr{C_1}1,1,2 Gn{Gn_2}2 Cr{C_3}3,3,4]')

    def test_resize(self):
        """
        Tests resizing of output layers.
        """
        rnn = vgsl.TorchVGSLModel('[1,1,0,48 Lbx10 Do O1c57]')
        rnn.resize_output(80)
        self.assertEqual(rnn.nn[-1].lin.out_features, 80)

    def test_del_resize(self):
        """
        Tests resizing of output layers with entry deletion.
        """
        rnn = vgsl.TorchVGSLModel('[1,1,0,48 Lbx10 Do O1c57]')
        rnn.resize_output(80, [2, 4, 5, 6, 7, 12, 25])
        self.assertEqual(rnn.nn[-1].lin.out_features, 80)

    def test_nested_serial_model(self):
        """
        Test the creation of a nested serial model.
        """
        net = vgsl.TorchVGSLModel('[1,48,0,1 Cr4,2,1,4,2 ([Cr4,2,1,1,1 Do Cr3,3,2,1,1] [Cr4,2,1,1,1 Cr3,3,2,1,1 Do]) S1(1x0)1,3 Lbx2 Do0.5 Lbx2]')
        self.assertIsInstance(net.nn[1], layers.MultiParamParallel)
        for x in net.nn[1].children():
            self.assertIsInstance(x, layers.MultiParamSequential)
            self.assertEqual(len(x), 3)

    def test_parallel_model_inequal(self):
        """
        Test proper raising of ValueError when parallel layers do not have the same output shape.
        """
        with raises(ValueError):
            net = vgsl.TorchVGSLModel('[1,48,0,1 Cr4,2,1,4,2 [Cr4,2,1,1,1 (Cr4,2,1,4,2 Cr3,3,2,1,1) S1(1x0)1,3 Lbx2 Do0.5] Lbx2]')

    def test_complex_serialization(self):
        """
        Test proper serialization and deserialization of a complex model.
        """
        net = vgsl.TorchVGSLModel('[1,48,0,1 Cr4,2,1,4,2 ([Cr4,2,1,1,1 Do Cr3,3,2,1,1] [Cr4,2,1,1,1 Cr3,3,2,1,1 Do]) S1(1x0)1,3 Lbx2 Do0.5 Lbx2]')



#
# Copyright 2015 Benjamin Kiessling
#           2014 Thomas M. Breuel
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
kraken.binarization
~~~~~~~~~~~~~~~~~~~

An adaptive binarization algorithm.
"""
import logging
import warnings
from typing import TYPE_CHECKING

import numpy as np
from scipy.ndimage import (affine_transform, binary_dilation, gaussian_filter,
                           percentile_filter)
from scipy.ndimage import zoom as _zoom

from kraken.lib.exceptions import KrakenInputException
from kraken.lib.util import array2pil, get_im_str, is_bitonal, pil2array

if TYPE_CHECKING:
    from PIL import Image


__all__ = ['nlbin']

logger = logging.getLogger(__name__)


def nlbin(im: 'Image.Image',
          threshold: float = 0.5,
          zoom: float = 0.5,
          escale: float = 1.0,
          border: float = 0.1,
          perc: int = 80,
          range: int = 20,
          low: int = 5,
          high: int = 90) -> 'Image.Image':
    """
    Performs binarization using non-linear processing.

    Args:
        im: Input image
        threshold:
        zoom: Zoom for background page estimation
        escale: Scale for estimating a mask over the text region
        border: Ignore this much of the border
        perc: Percentage for filters
        range: Range for filters
        low: Percentile for black estimation
        high: Percentile for white estimation

    Returns:
        PIL.Image.Image containing the binarized image

    Raises:
        KrakenInputException: When trying to binarize an empty image.
    """
    im_str = get_im_str(im)
    logger.info(f'Binarizing {im_str}')
    if is_bitonal(im):
        logger.info(f'Skipping binarization because {im_str} is bitonal.')
        return im
    # convert to grayscale first
    logger.debug(f'Converting {im_str} to grayscale')
    im = im.convert('L')
    raw = pil2array(im)
    logger.debug('Scaling and normalizing')
    # rescale image to between -1 or 0 and 1
    raw = raw/float(np.iinfo(raw.dtype).max)
    # perform image normalization
    if np.amax(raw) == np.amin(raw):
        logger.warning(f'Trying to binarize empty image {im_str}')
        raise KrakenInputException('Image is empty')
    image = raw-np.amin(raw)
    image /= np.amax(image)

    logger.debug('Interpolation and percentile filtering')
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', UserWarning)
        m = _zoom(image, zoom)
        m = percentile_filter(m, perc, size=(range, 2))
        m = percentile_filter(m, perc, size=(2, range))
        mh, mw = m.shape
        oh, ow = image.shape
        scale = np.diag([mh * 1.0/oh, mw * 1.0/ow])
        m = affine_transform(m, scale, output_shape=image.shape)
    w, h = np.minimum(np.array(image.shape), np.array(m.shape))
    flat = np.clip(image[:w, :h]-m[:w, :h]+1, 0, 1)

    # estimate low and high thresholds
    d0, d1 = flat.shape
    o0, o1 = int(border*d0), int(border*d1)
    est = flat[o0:d0-o0, o1:d1-o1]
    logger.debug('Threshold estimates {}'.format(est))
    # by default, we use only regions that contain
    # significant variance; this makes the percentile
    # based low and high estimates more reliable
    logger.debug('Refine estimates')
    v = est-gaussian_filter(est, escale*20.0)
    v = gaussian_filter(v**2, escale*20.0)**0.5
    v = (v > 0.3*np.amax(v))
    v = binary_dilation(v, structure=np.ones((int(escale * 50), 1)))
    v = binary_dilation(v, structure=np.ones((1, int(escale * 50))))
    est = est[v]
    lo = np.percentile(est.ravel(), low)
    hi = np.percentile(est.ravel(), high)
    flat -= lo
    flat /= (hi-lo)
    flat = np.clip(flat, 0, 1)
    logger.debug(f'Thresholding at {threshold}')
    bin = np.array(255*(flat > threshold), 'B')
    return array2pil(bin)



# -*- coding: utf-8 -*-

import unittest
from pathlib import Path

from kraken.align import forced_align
from kraken.lib import xml

thisfile = Path(__file__).resolve().parent
resources = thisfile / 'resources'

class TestKrakenAlign(unittest.TestCase):
    """
    Tests for the forced alignment module
    """
    def setUp(self):
        self.doc = resources / '170025120000003,0074.xml'
        self.bls = xml.XMLPage(self.doc).to_container()

    def test_forced_align_simple(self):
        """
        Simple alignment test.
        """
        pass



#!/usr/bin/env python
# 2020, Bruce Robertson
# Master file at https://github.com/brobertson/Lace2-tools/blob/master/normalize_hocr.py
import argparse
import html
import os
import sys
from statistics import mean

from lxml import etree
from PIL import Image


# a custom exception to indicate when a page or other element doesn't
# have a bounding box where we would expect it.
class BboxError(Exception):
    pass


# parse the arguments
parser = argparse.ArgumentParser(description='''Convert kraken hocr output so
                                 that word bounding boxes are very likely to enclose the words, plus some space.
                                 This removes all spans of class ocrx_word that
                                 have single space text content. Its output is
                                 namespaced XHTML.''')
parser.add_argument('--inputDir', help='Path to directory where source files are found', required=True)
parser.add_argument('--outputDir', help='Path to directory where output is stored', required=True)
parser.add_argument('--imageDir', help='Path to directory where images corresponding to the html files are stored.', required=False)
parser.add_argument('-c', '--confidenceSummary', default=False, action="store_true",
                    help="store summaries of word confidence in xhtml data- "
                         "attributes and cut all material after the first ; from the "
                         "word span title attribute, making their mouseover popups "
                         "less obtrusive.")
parser.add_argument('-f', '--fixBigWordSpans', default=False, action="store_true",
                    help="fix word_span elements whose bbox area is greater "
                         "than a sixth of the whole page area by assigning them the "
                         "bbox of the previous word.")
parser.add_argument('-s', '--shareSpaceSpans', default=False, action="store_true",
                    help="normalize hocr output from kraken, which assigns a "
                         "word to every space and gives it a bbox. This removes those "
                         "space words and assigns their area to the words on either "
                         "side, with some space in between, generating output more "
                         "like Ocropus and tesseract.")
parser.add_argument("-v", "--verbose", help="increase output verbosity", default=False, action="store_true")
args = parser.parse_args()


def get_bbox_val(span, position):
    try:
        parts = html.unescape(span.get('title')).split(';')
        bbox_string = ""
        for part in parts:
            part = part.strip()
            if part.startswith('bbox'):
                bbox_string = part
        if bbox_string == "":
            if (args.verbose):
                print("couldn't find the bbox part!")
        return int(bbox_string.split(' ')[position + 1])
    except Exception as e:
        # print("Exception getting title element on span {}".format(etree.tostring(span)))
        if (args.verbose):
            print(e)
            print("... therefore raising BboxError")
        raise BboxError


def get_bbox_area(span):
    try:
        width = get_bbox_val(span, 2) - get_bbox_val(span, 0)
        height = get_bbox_val(span, 3) - get_bbox_val(span, 1)
        area = width * height
        if (args.verbose):
            print("this element's area is " + str(area))
        return area
    except Exception:
        # print("Exception getting area on span  {}".format(etree.tostring(span)))
        raise


def set_bbox_value(span, position, val):
    try:
        parts = span.get('title').split(';')
    except Exception:
        print("Exception getting title element on span id {}.".format(span.get('id')))
        raise BboxError
    bbox_parts = parts[0].split(' ')
    bbox_parts[position + 1] = str(val)
    bbox_out = ' '.join(bbox_parts)
    parts[0] = bbox_out
    parts_out = ';'.join(parts)
    span.set('title', parts_out)


def share_space_spans(treeIn):
    right_max_fudge_factor = 7
    left_max_fudge_factor = 5
    space_spans = treeIn.xpath("//html:span[@class='ocrx_word'][text()=' ']", namespaces={'html': "http://www.w3.org/1999/xhtml"})
    # print('space spans: {}'.format(len(space_spans)))
    for space_span in space_spans:
        try:
            previous_span = space_span.getprevious()
        except Exception as e:
            print("Exception on parsing previous span with space id {}".format(space_span.get('id')))
            print(e)
            raise
        try:
            next_span = space_span.getnext()
        except Exception as e:
            print("Exception on parsing next span with space id {}".format(space_span.get('id')))
            print(e)
            raise
        # check that we have both
        if previous_span and next_span:
            # this means that there is both a previous and a next
            if (args.verbose):
                print("***")
                print("space_span title: {}".format(space_span.get('title')))
                print("previous span title: {}".format(previous_span.get('title')))
                print("next span title: {}".format(next_span.get('title')))
            left_pos = get_bbox_val(previous_span, 2)
            right_pos = get_bbox_val(next_span, 0)
            middle = int((left_pos + right_pos) / 2)
            third = int((right_pos - left_pos) / 3)
            left_fudge = min(third, left_max_fudge_factor)
            right_fudge = min(third, right_max_fudge_factor)
            if (args.verbose):
                print("left side: {0}; right side: {1}; middle: {2}".format(left_pos, right_pos, middle))
            set_bbox_value(previous_span, 2, left_pos + left_fudge)
            set_bbox_value(next_span, 0, right_pos - right_fudge)
            if (args.verbose):
                print(previous_span.text)
                print("previous_span new title: {}".format(previous_span.get('title')))
                print("next_span new title: {}".format(next_span.get('title')))
        # now remove the space span, no matter what
        space_span.getparent().remove(space_span)


def confidence_summary(treeIn):
    word_spans = treeIn.xpath("//html:span[@class='ocrx_word']", namespaces={'html': "http://www.w3.org/1999/xhtml"})
    for word_span in word_spans:
        try:
            # this gets the confidence values for each letter and represents them as a string list
            word_data = word_span.get('title').split(';')
            confs_string = word_data[1].split(' ')[2:]
            bbox_only = word_data[0]
            # convert to floats for math operations
            confs = [float(i) for i in confs_string]
            minimum = round(min(confs), 2)
            average = round(mean(confs), 2)
            # add attributes with these summary values
            word_span.set('data-min-confidence', str(minimum))
            word_span.set('data-average-confidence', str(average))
            word_span.set('title', bbox_only)
        except Exception:
            # there's not much to do if this goes wrong
            pass


def push_edge_spans_to_borders_of_line(treeIn):
    first_spans = treeIn.xpath("//html:span[@class='ocr_line']/html:span[@class='ocrx_word'][1]", namespaces={'html': "http://www.w3.org/1999/xhtml"})
    for span in first_spans:
        if (args.verbose):
            print("first span title: {}".format(span.get('title')))
        parent = span.getparent()
        line_l_edge = get_bbox_val(parent, 0)
        line_l_edge = line_l_edge + 1
        if (args.verbose):
            print("line_l_edge {}".format(line_l_edge))
        set_bbox_value(span, 0, line_l_edge)
    last_spans = treeIn.xpath("//html:span[@class='ocr_line']/html:span[@class='ocrx_word'][last()]", namespaces={'html': "http://www.w3.org/1999/xhtml"})
    for span in last_spans:
        parent = span.getparent()
        line_r_edge = get_bbox_val(parent, 2)
        line_r_edge = line_r_edge - 1
        set_bbox_value(span, 2, line_r_edge)


def fix_word_span_area(treeIn):
    word_spans = treeIn.xpath("//html:span[@class='ocrx_word'] | //html:span[@class='ocr_word']", namespaces={'html': "http://www.w3.org/1999/xhtml"})
    image_area = get_bbox_area(treeIn.xpath("//html:div[@class='ocr_page'][1]", namespaces={'html': "http://www.w3.org/1999/xhtml"})[0])
    # print("image area: {}".format(image_area))
    for span in word_spans:
        area = get_bbox_area(span)
        # print("word area:",area)
        if (area > image_area / 6):
            original_area = span.get('title')
            my_next = span.getnext()
            my_previous = span.getprevious()
            if (my_next):
                span.set('title', my_next.get('title'))
            elif (my_previous):
                span.set('title', my_previous.get('title'))
            else:
                span.set('title', span.getparent().get('title'))
            print("big word area, changing title attribute {}".format(original_area))
            print(etree.tostring(span))
            # span.getparent().remove(span)


def clean_ocr_page_title(xhtml, file_name):
    ocr_page = xhtml.xpath("//html:div[@class='ocr_page'][1]", namespaces={'html': "http://www.w3.org/1999/xhtml"})[0]
    # print(ocr_page)
    ocr_page_title = ocr_page.get('title')
    # print(ocr_page_title)
    sections = ocr_page_title.split(';')
    # print(sections)
    new_sections = "image " + (file_name.rsplit('.', 1)[0] + '.png') + "; " + sections[0]
    # print(new_sections)
    ocr_page.set('title', new_sections)
    return xhtml


def rewrite_ocr_page_title(xhtml, file_name, image_x, image_y):
    ocr_page = xhtml.xpath("//html:div[@class='ocr_page'][1]", namespaces={'html': "http://www.w3.org/1999/xhtml"})[0]
    image_file_name = (file_name.rsplit('.', 1)[0] + '.png')
    image_path = os.path.join(args.imageDir, image_file_name)
    image = Image.open(image_path)
    image_x, image_y = image.size
    new_title = "bbox 0 0 " + str(image_x) + " " + str(image_y) + ";image " + image_file_name
    ocr_page.set('title', new_title)
    return xhtml


if not (os.path.isdir(args.inputDir)):
    print('Input directory "' + args.inputDir + '" does not exist.\n\tExiting ...')
    sys.exit(1)

# Create the output directory if it doesn't exist
try:
    if not os.path.exists(args.outputDir):
        os.makedirs(args.outputDir, exist_ok=True)
except Exception as e:
    print("Error on creating output directory '" + args.outputDir +
          "':\n\t" + str(e) + "\n\tExiting ...")
    sys.exit(1)

if (args.verbose):
    print("Input dir:", args.inputDir)
    print("Output dir:", args.outputDir)

# everything looks good. Let's loop over the html files in inputDir
xslt_to_xhtml = etree.XML('''\
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform" version="1.0"
   xmlns:html='http://www.w3.org/1999/xhtml'>

   <xsl:template match="*">
    <xsl:element name="html:{local-name(.)}">
      <xsl:apply-templates select="@*|*|text()"/>
       </xsl:element>
       </xsl:template>

       <xsl:template match="@*">
         <xsl:attribute name="{name(.)}"><xsl:value-of
         select="."/></xsl:attribute>
         </xsl:template>

         </xsl:stylesheet>''')
transform_to_xhtml = etree.XSLT(xslt_to_xhtml)

EXTENSIONS = ('.hocr', '.html', '.htm')
XHTML_NAMESPACE = "http://www.w3.org/1999/xhtml"
for root, dirs, files in os.walk(args.inputDir):
    for file_name in files:
        if file_name.endswith(EXTENSIONS):
            print(file_name)
            with open(os.path.join(args.inputDir, file_name)) as file:  # Use file to refer to the file
                try:
                    tree = etree.parse(file)
                    find_xhtml_body = etree.ETXPath("//{%s}body" % XHTML_NAMESPACE)
                    results = find_xhtml_body(tree)
                    xhtml = transform_to_xhtml(tree)
                    if (args.fixBigWordSpans):
                        try:
                            fix_word_span_area(xhtml)
                        except BboxError:
                            if (args.verbose):
                                print("we get a bbox error while trying to fix word span areas on " +
                                      file_name + " so we'll try to rewrite the page bbox using imageDir")
                            try:
                                # todo check that args.imageDir is set
                                if not (os.path.isdir(args.imageDir)):
                                    print("To correct these files, the imageDir argument has to be set to a valid directory.")
                                    exit(1)
                                # remove '.html' and add '.png'
                                image_file_name = file_name[:-5] + '.png'
                                image_path = os.path.join(args.imageDir, image_file_name)
                                image = Image.open(image_path)
                                image_x, image_y = image.size
                                rewrite_ocr_page_title(xhtml, file_name, image_x, image_y)
                                fix_word_span_area(xhtml)
                            except Exception:
                                print("something went wrong trying to fix the page bbox on this file. Aborting.")
                                raise
                    clean_ocr_page_title(xhtml, file_name)
                    if (args.shareSpaceSpans):
                        share_space_spans(xhtml)
                    if (args.confidenceSummary):
                        confidence_summary(xhtml)
                    # push_edge_spans_to_borders_of_line(xhtml)
                    xhtml.write(os.path.join(args.outputDir, file_name), pretty_print=True, xml_declaration=True, encoding="utf-8")
                except Exception as e:
                    print("This exception was thrown on file {}".format(file_name))
                    print(e)



#
# Copyright 2023 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
Pytorch-lightning modules for reading order training.

Adapted from:
"""
import logging
from dataclasses import dataclass, field
from typing import (TYPE_CHECKING, Any, Dict, List, Literal, Optional,
                    Sequence, Union)

import numpy as np
import lightning as L
import torch
import torch.nn.functional as F
from lightning.pytorch.callbacks import EarlyStopping, LearningRateMonitor
from torch.optim import lr_scheduler
from torch.utils.data import DataLoader, Subset

from kraken.lib import default_specs
from kraken.lib.dataset import PageWiseROSet, PairWiseROSet
from kraken.lib.ro.layers import MLP
from kraken.lib.segmentation import _greedy_order_decoder
from kraken.lib.train import _configure_optimizer_and_lr_scheduler

if TYPE_CHECKING:
    from os import PathLike

    from torch.nn import Module

logger = logging.getLogger(__name__)


@dataclass
class DummyVGSLModel:
    hyper_params: Dict[str, int] = field(default_factory=dict)
    user_metadata: Dict[str, List] = field(default_factory=dict)
    one_channel_mode: Literal['1', 'L'] = '1'
    ptl_module: 'Module' = None
    model_type: str = 'unknown'

    def __post_init__(self):
        self.hyper_params: Dict[str, int] = {'completed_epochs': 0}
        self.user_metadata: Dict[str, List] = {'accuracy': [], 'metrics': []}

    def save_model(self, filename):
        self.ptl_module.save_checkpoint(filename)


def spearman_footrule_distance(s, t):
    return (s - t).abs().sum() / (0.5 * (len(s) ** 2 - (len(s) % 2)))


class ROModel(L.LightningModule):
    def __init__(self,
                 hyper_params: Dict[str, Any] = None,
                 output: str = 'model',
                 training_data: Union[Sequence[Union['PathLike', str]], Sequence[Dict[str, Any]]] = None,
                 evaluation_data: Optional[Union[Sequence[Union['PathLike', str]], Sequence[Dict[str, Any]]]] = None,
                 partition: Optional[float] = 0.9,
                 num_workers: int = 1,
                 format_type: Literal['alto', 'page', 'xml'] = 'xml',
                 load_hyper_parameters: bool = False,
                 level: Literal['baselines', 'regions'] = 'baselines',
                 reading_order: Optional[str] = None):
        """
        A LightningModule encapsulating the unsupervised pretraining setup for
        a text recognition model.

        Setup parameters (load, training_data, evaluation_data, ....) are
        named, model hyperparameters (everything in
        `kraken.lib.default_specs.RECOGNITION_HYPER_PARAMS`) are in in the
        `hyper_params` argument.

        Args:
            hyper_params (dict): Hyperparameter dictionary containing all fields
                                 from
                                 kraken.lib.default_specs.RECOGNITION_PRETRAIN_HYPER_PARAMS
            **kwargs: Setup parameters, i.e. CLI parameters of the train() command.
        """
        super().__init__()
        self.hyper_params = default_specs.READING_ORDER_HYPER_PARAMS
        if hyper_params:
            self.hyper_params.update(hyper_params)

        if not evaluation_data:
            np.random.shuffle(training_data)
            training_data = training_data[:int(partition*len(training_data))]
            evaluation_data = training_data[int(partition*len(training_data)):]
        train_set = PairWiseROSet(training_data,
                                  mode=format_type,
                                  level=level,
                                  ro_id=reading_order)
        self.train_set = Subset(train_set, range(len(train_set)))
        self.class_mapping = train_set.class_mapping
        val_set = PageWiseROSet(evaluation_data,
                                mode=format_type,
                                class_mapping=train_set.class_mapping,
                                level=level,
                                ro_id=reading_order)
        self.val_set = Subset(val_set, range(len(val_set)))

        if len(self.train_set) == 0 or len(self.val_set) == 0:
            raise ValueError('No valid training data was provided to the train '
                             'command. Please add valid XML, line, or binary data.')

        logger.info(f'Training set {len(self.train_set)} lines, validation set '
                    f'{len(self.val_set)} lines')

        self.output = output
        self.criterion = torch.nn.BCEWithLogitsLoss()

        self.num_workers = num_workers

        self.best_epoch = -1
        self.best_metric = torch.inf

        logger.info('Creating new RO model')
        self.ro_net = MLP(train_set.get_feature_dim(), train_set.get_feature_dim() * 2)

        if 'file_system' in torch.multiprocessing.get_all_sharing_strategies():
            logger.debug('Setting multiprocessing tensor sharing strategy to file_system')
            torch.multiprocessing.set_sharing_strategy('file_system')

        self.nn = DummyVGSLModel(ptl_module=self)

        self.val_losses = []
        self.val_spearman = []

        self.save_hyperparameters()

    def forward(self, x):
        return F.sigmoid(self.ro_net(x))

    def validation_step(self, batch, batch_idx):
        xs, ys, num_lines = batch['sample'], batch['target'], batch['num_lines']
        logits = self.ro_net(xs).squeeze()
        yhat = F.sigmoid(logits)
        order = torch.zeros((num_lines, num_lines))
        idx = 0
        for i in range(num_lines):
            for j in range(num_lines):
                if i != j:
                    order[i, j] = yhat[idx]
                    idx += 1
        path = _greedy_order_decoder(order)
        spearman_dist = spearman_footrule_distance(torch.tensor(range(num_lines)), path)
        self.log('val_spearman', spearman_dist)
        loss = self.criterion(logits, ys.squeeze())
        self.val_losses.append(loss.cpu())
        self.val_spearman.append(spearman_dist.cpu())

    def on_validation_epoch_end(self):
        if not self.trainer.sanity_checking:
            val_metric = np.mean(self.val_spearman)
            val_loss = np.mean(self.val_losses)

            if val_metric < self.best_metric:
                logger.debug(f'Updating best metric from {self.best_metric} ({self.best_epoch}) to {val_metric} ({self.current_epoch})')
                self.best_epoch = self.current_epoch
                self.best_metric = val_metric
            logger.info(f'validation run: val_spearman {val_metric} val_loss {val_loss}')
            self.log('val_spearman', val_metric, on_step=False, on_epoch=True, prog_bar=True, logger=True)
            self.log('val_metric', val_metric, on_step=False, on_epoch=True, prog_bar=False, logger=True)
            self.log('val_loss', val_loss, on_step=False, on_epoch=True, prog_bar=True, logger=True)

        self.val_spearman.clear()
        self.val_losses.clear()

    def training_step(self, batch, batch_idx):
        x, y = batch['sample'], batch['target']
        logits = self.ro_net(x)
        loss = self.criterion(logits.squeeze(), y)
        self.log('loss', loss)
        return loss

    def train_dataloader(self):
        return DataLoader(self.train_set,
                          batch_size=self.hyper_params['batch_size'],
                          num_workers=self.num_workers,
                          pin_memory=True)

    def val_dataloader(self):
        return DataLoader(self.val_set,
                          batch_size=1,
                          num_workers=self.num_workers,
                          pin_memory=True)

    def save_checkpoint(self, filename):
        self.trainer.save_checkpoint(filename)

    def configure_callbacks(self):
        callbacks = []
        if self.hparams.hyper_params['quit'] == 'early':
            callbacks.append(EarlyStopping(monitor='val_metric',
                                           mode='min',
                                           patience=self.hparams.hyper_params['lag'],
                                           stopping_threshold=0.0))
        if self.hparams.hyper_params['pl_logger']:
            callbacks.append(LearningRateMonitor(logging_interval='step'))
        return callbacks

    # configuration of optimizers and learning rate schedulers
    # --------------------------------------------------------
    #
    # All schedulers are created internally with a frequency of step to enable
    # batch-wise learning rate warmup. In lr_scheduler_step() calls to the
    # scheduler are then only performed at the end of the epoch.
    def configure_optimizers(self):
        return _configure_optimizer_and_lr_scheduler(self.hparams.hyper_params,
                                                     self.ro_net.parameters(),
                                                     len_train_set=len(self.train_set),
                                                     loss_tracking_mode='min')

    def optimizer_step(self, epoch, batch_idx, optimizer, optimizer_closure):
        # update params
        optimizer.step(closure=optimizer_closure)

        # linear warmup between 0 and the initial learning rate `lrate` in `warmup`
        # steps.
        if self.hparams.hyper_params['warmup'] and self.trainer.global_step < self.hparams.hyper_params['warmup']:
            lr_scale = min(1.0, float(self.trainer.global_step + 1) / self.hparams.hyper_params['warmup'])
            for pg in optimizer.param_groups:
                pg["lr"] = lr_scale * self.hparams.hyper_params['lrate']

    def lr_scheduler_step(self, scheduler, metric):
        if not self.hparams.hyper_params['warmup'] or self.trainer.global_step >= self.hparams.hyper_params['warmup']:
            # step OneCycleLR each batch if not in warmup phase
            if isinstance(scheduler, lr_scheduler.OneCycleLR):
                scheduler.step()
            # step every other scheduler epoch-wise
            elif self.trainer.is_last_batch:
                if metric is None:
                    scheduler.step()
                else:
                    scheduler.step(metric)



#
# Copyright 2023 Benjamin Kiessling
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express
# or implied. See the License for the specific language governing
# permissions and limitations under the License.
"""
Tools for trainable reading order.
"""

from .model import ROModel  # NOQA



# -*- coding: utf-8 -*-
import json
import tempfile
import unittest
from pathlib import Path

import numpy as np
from pytest import raises

from kraken.lib import xml

thisfile = Path(__file__).resolve().parent
resources = thisfile / 'resources'

class TestXMLParser(unittest.TestCase):
    """
    Tests XML (ALTO/PAGE) parsing
    """
    def setUp(self):
        self.page_doc = resources / 'cPAS-2000.xml'
        self.alto_doc = resources / 'bsb00084914_00007.xml'

    def test_page_parsing(self):
        """
        Test parsing of PAGE XML files with reading order.
        """
        doc = xml.XMLPage(self.page_doc, filetype='page')
        self.assertEqual(len(doc.get_sorted_lines()), 97)
        self.assertEqual(len([item for x in doc.regions.values() for item in x]), 4)

    def test_alto_parsing(self):
        """
        Test parsing of ALTO XML files with reading order.
        """
        doc = xml.XMLPage(self.alto_doc, filetype='alto')

    def test_auto_parsing(self):
        """
        Test parsing of PAGE and ALTO XML files with auto-format determination.
        """
        doc = xml.XMLPage(self.page_doc, filetype='xml')
        self.assertEqual(doc.filetype, 'page')
        doc = xml.XMLPage(self.alto_doc, filetype='xml')
        self.assertEqual(doc.filetype, 'alto')

    def test_failure_page_alto_parsing(self):
        """
        Test that parsing ALTO files with PAGE as format fails.
        """
        with raises(ValueError):
            xml.XMLPage(self.alto_doc, filetype='page')

    def test_failure_alto_page_parsing(self):
        """
        Test that parsing PAGE files with ALTO as format fails.
        """
        with raises(ValueError):
            xml.XMLPage(self.page_doc, filetype='alto')

